# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the models.py module of the skill."""

# pylint: skip-file

import builtins
import json
import logging
from collections import OrderedDict
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from tempfile import TemporaryDirectory
from time import sleep
from typing import Any, Dict, List, Optional, Tuple, Type, cast
from unittest import mock
from unittest.mock import MagicMock

import pytest
from aea.exceptions import AEAEnforceError
from typing_extensions import Literal, TypedDict

from packages.valory.skills.abstract_round_abci.base import (
    AbstractRound,
    BaseSynchronizedData,
    ROUND_COUNT_DEFAULT,
)
from packages.valory.skills.abstract_round_abci.models import (
    ApiSpecs,
    BaseParams,
    BenchmarkTool,
    DEFAULT_BACKOFF_FACTOR,
    GenesisBlock,
    GenesisConfig,
    GenesisConsensusParams,
    GenesisEvidence,
    GenesisValidator,
    MIN_RESET_PAUSE_DURATION,
    NUMBER_OF_RETRIES,
    Requests,
)
from packages.valory.skills.abstract_round_abci.models import (
    SharedState as BaseSharedState,
)
from packages.valory.skills.abstract_round_abci.models import (
    TendermintRecoveryParams,
    _MetaSharedState,
    check_type,
)
from packages.valory.skills.abstract_round_abci.test_tools.abci_app import AbciAppTest
from packages.valory.skills.abstract_round_abci.tests.conftest import (
    irrelevant_genesis_config,
)


BASE_DUMMY_SPECS_CONFIG = dict(
    name="dummy",
    skill_context=MagicMock(),
    url="http://dummy",
    api_id="api_id",
    method="GET",
    headers=OrderedDict([("Dummy-Header", "dummy_value")]),
    parameters=OrderedDict([("Dummy-Param", "dummy_param")]),
)

BASE_DUMMY_PARAMS = dict(
    name="",
    skill_context=MagicMock(is_abstract_component=True),
    setup={},
    tendermint_url="",
    max_healthcheck=1,
    round_timeout_seconds=1.0,
    sleep_time=1,
    retry_timeout=1,
    retry_attempts=1,
    reset_pause_duration=MIN_RESET_PAUSE_DURATION,
    drand_public_key="",
    tendermint_com_url="",
    reset_tendermint_after=1,
    service_id="abstract_round_abci",
    service_registry_address="0xa51c1fc2f0d1a1b8494ed1fe312d7c3a78ed91c0",
    keeper_timeout=1.0,
    tendermint_check_sleep_delay=3,
    tendermint_max_retries=5,
    cleanup_history_depth=0,
    genesis_config=irrelevant_genesis_config,
    cleanup_history_depth_current=None,
    request_timeout=0.0,
    request_retry_delay=0.0,
    tx_timeout=0.0,
    max_attempts=0,
    on_chain_service_id=None,
    share_tm_config_on_startup=False,
    tendermint_p2p_url="",
    use_termination=False,
)


class TestApiSpecsModel:
    """Test ApiSpecsModel."""

    api_specs: ApiSpecs

    def setup(
        self,
    ) -> None:
        """Setup test."""

        self.api_specs = ApiSpecs(
            **BASE_DUMMY_SPECS_CONFIG,
            response_key="value",
            response_index=0,
            response_type="float",
            error_key="error",
            error_index=None,
            error_type="str",
            error_data="error text",
        )

    def test_init(
        self,
    ) -> None:
        """Test initialization."""

        # test ensure method.
        with pytest.raises(
            AEAEnforceError,
            match="'url' of type '<class 'str'>' required, but it is not set in `models.params.args` of `skill.yaml` of",
        ):
            _ = ApiSpecs(
                name="dummy",
                skill_context=MagicMock(),
            )

        assert self.api_specs.retries_info.backoff_factor == DEFAULT_BACKOFF_FACTOR
        assert self.api_specs.retries_info.retries == NUMBER_OF_RETRIES
        assert self.api_specs.retries_info.retries_attempted == 0

        assert self.api_specs.url == "http://dummy"
        assert self.api_specs.api_id == "api_id"
        assert self.api_specs.method == "GET"
        assert self.api_specs.headers == {"Dummy-Header": "dummy_value"}
        assert self.api_specs.parameters == {"Dummy-Param": "dummy_param"}
        assert self.api_specs.response_info.response_key == "value"
        assert self.api_specs.response_info.response_index == 0
        assert self.api_specs.response_info.response_type == "float"
        assert self.api_specs.response_info.error_key == "error"
        assert self.api_specs.response_info.error_index is None
        assert self.api_specs.response_info.error_type == "str"
        assert self.api_specs.response_info.error_data is None

    @pytest.mark.parametrize("retries", range(10))
    def test_suggested_sleep_time(self, retries: int) -> None:
        """Test `suggested_sleep_time`"""
        self.api_specs.retries_info.retries_attempted = retries
        assert (
            self.api_specs.retries_info.suggested_sleep_time
            == DEFAULT_BACKOFF_FACTOR**retries
        )

    def test_retries(
        self,
    ) -> None:
        """Tests for retries."""

        self.api_specs.increment_retries()
        assert self.api_specs.retries_info.retries_attempted == 1
        assert not self.api_specs.is_retries_exceeded()

        for _ in range(NUMBER_OF_RETRIES):
            self.api_specs.increment_retries()
        assert self.api_specs.is_retries_exceeded()
        self.api_specs.reset_retries()
        assert self.api_specs.retries_info.retries_attempted == 0

    def test_get_spec(
        self,
    ) -> None:
        """Test get_spec method."""

        actual_specs = {
            "url": "http://dummy",
            "method": "GET",
            "headers": {"Dummy-Header": "dummy_value"},
            "parameters": {"Dummy-Param": "dummy_param"},
        }

        specs = self.api_specs.get_spec()
        assert all([key in specs for key in actual_specs.keys()])
        assert all([specs[key] == actual_specs[key] for key in actual_specs])

    @pytest.mark.parametrize(
        "api_specs_config, message, expected_res, expected_error",
        (
            (
                dict(
                    **BASE_DUMMY_SPECS_CONFIG,
                    response_key="value",
                    response_index=None,
                    response_type="float",
                    error_key=None,
                    error_index=None,
                    error_data=None,
                ),
                MagicMock(body=b'{"value": "10.232"}'),
                10.232,
                None,
            ),
            (
                dict(
                    **BASE_DUMMY_SPECS_CONFIG,
                    response_key="test:response:key",
                    response_index=2,
                    response_type="dict",
                    error_key="error:key",
                    error_index=3,
                    error_type="str",
                    error_data=None,
                ),
                MagicMock(
                    body=b'{"test": {"response": {"key": ["does_not_matter", "does_not_matter", {"this": "matters"}]}}}'
                ),
                {"this": "matters"},
                None,
            ),
            (
                dict(
                    **BASE_DUMMY_SPECS_CONFIG,
                    response_key="test:response:key",
                    response_index=2,
                    error_key="error:key",
                    error_index=3,
                    error_type="str",
                    error_data=None,
                ),
                MagicMock(body=b'{"cannot be parsed'),
                None,
                None,
            ),
            (
                dict(
                    **BASE_DUMMY_SPECS_CONFIG,
                    response_key="test:response:key",
                    response_index=2,
                    error_key="error:key",
                    error_index=3,
                    error_type="str",
                    error_data=None,
                ),
                MagicMock(
                    # the null will raise `TypeError` and we test that it is handled
                    body=b'{"test": {"response": {"key": ["does_not_matter", "does_not_matter", null]}}}'
                ),
                "None",
                None,
            ),
            (
                dict(
                    **BASE_DUMMY_SPECS_CONFIG,
                    response_key="test:response:key",
                    response_index=2,  # this will raise `IndexError` and we test that it is handled
                    error_key="error:key",
                    error_index=3,
                    error_type="str",
                    error_data=None,
                ),
                MagicMock(
                    body=b'{"test": {"response": {"key": ["does_not_matter", "does_not_matter"]}}}'
                ),
                None,
                None,
            ),
            (
                dict(
                    **BASE_DUMMY_SPECS_CONFIG,
                    response_key="test:response:key",  # this will raise `KeyError` and we test that it is handled
                    response_index=2,
                    error_key="error:key",
                    error_index=3,
                    error_type="str",
                    error_data=None,
                ),
                MagicMock(
                    body=b'{"test": {"response": {"key_does_not_match": ["does_not_matter", "does_not_matter"]}}}'
                ),
                None,
                None,
            ),
            (
                dict(
                    **BASE_DUMMY_SPECS_CONFIG,
                    response_key="test:response:key",
                    response_index=2,
                    error_key="error:key",
                    error_index=3,
                    error_type="str",
                    error_data=None,
                ),
                MagicMock(
                    body=b'{"test": {"response": {"key_does_not_match": ["does_not_matter", "does_not_matter"]}}, '
                    b'"error": {"key": [0, 1, 2, "test that the error is being parsed correctly"]}}'
                ),
                None,
                "test that the error is being parsed correctly",
            ),
        ),
    )
    def test_process_response(
        self,
        api_specs_config: dict,
        message: MagicMock,
        expected_res: Any,
        expected_error: Any,
    ) -> None:
        """Test `process_response` method."""
        api_specs = ApiSpecs(**api_specs_config)
        actual = api_specs.process_response(message)
        assert actual == expected_res
        response_type = api_specs_config.get("response_type", None)
        if response_type is not None:
            assert type(actual) == getattr(builtins, response_type)
        assert api_specs.response_info.error_data == expected_error

    def test_attribute_manipulation(self) -> None:
        """Test manipulating the attributes."""
        with pytest.raises(AttributeError, match="This object is frozen!"):
            del self.api_specs.url

        with pytest.raises(AttributeError, match="This object is frozen!"):
            self.api_specs.url = ""

        self.api_specs.__dict__["_frozen"] = False
        self.api_specs.url = ""
        del self.api_specs.url


class ConcreteRound(AbstractRound):
    """A ConcreteRoundA for testing purposes."""

    synchronized_data_class = MagicMock()
    payload_attribute = MagicMock()
    payload_class = MagicMock()

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """Handle the end of the block."""


class SharedState(BaseSharedState):
    """Shared State for testing purposes."""

    abci_app_cls = AbciAppTest


class TestSharedState:
    """Test SharedState(Model) class."""

    def test_initialization(self, *_: Any) -> None:
        """Test the initialization of the shared state."""
        SharedState(name="", skill_context=MagicMock())

    @staticmethod
    def dummy_state_setup(shared_state: SharedState) -> None:
        """Setup a shared state instance with dummy params."""
        shared_state.context.params.setup_params = {
            "test": [],
            "all_participants": list(range(4)),
        }
        shared_state.setup()

    def test_setup(self, *_: Any) -> None:
        """Test setup method."""
        shared_state = SharedState(
            name="", skill_context=MagicMock(is_abstract_component=False)
        )
        assert shared_state.initial_tm_configs == {}
        self.dummy_state_setup(shared_state)
        assert shared_state.initial_tm_configs == {i: None for i in range(4)}

    @pytest.mark.parametrize("self_idx", (range(4)))
    def test_acn_container(self, self_idx: int) -> None:
        """Test the `acn_container` method."""

        shared_state = SharedState(
            name="", skill_context=MagicMock(agent_address=self_idx)
        )
        self.dummy_state_setup(shared_state)
        expected = {i: None for i in range(4) if i != self_idx}
        assert shared_state.acn_container() == expected

    def test_synchronized_data_negative_not_available(self, *_: Any) -> None:
        """Test 'synchronized_data' property getter, negative case (not available)."""
        shared_state = SharedState(name="", skill_context=MagicMock())
        with pytest.raises(ValueError, match="round sequence not available"):
            shared_state.synchronized_data

    def test_synchronized_data_positive(self, *_: Any) -> None:
        """Test 'synchronized_data' property getter, negative case (not available)."""
        shared_state = SharedState(name="", skill_context=MagicMock())
        shared_state.context.params.setup_params = {
            "test": [],
            "all_participants": [["0x0"]],
        }
        shared_state.setup()
        shared_state.round_sequence.abci_app._round_results = [MagicMock()]
        shared_state.synchronized_data

    def test_synchronized_data_db(self, *_: Any) -> None:
        """Test 'synchronized_data' AbciAppDB."""
        shared_state = SharedState(name="", skill_context=MagicMock())
        with mock.patch.object(shared_state.context, "params") as mock_params:
            mock_params.setup_params = {
                "safe_contract_address": "0xsafe",
                "oracle_contract_address": "0xoracle",
                "all_participants": "0x0",
            }
            shared_state.setup()
            assert (
                shared_state.synchronized_data.db.get_strict("safe_contract_address")
                == "0xsafe"
            )
            assert (
                shared_state.synchronized_data.db.get_strict("oracle_contract_address")
                == "0xoracle"
            )

    @pytest.mark.parametrize(
        "address_to_acn_deliverable, n_participants, expected",
        (
            ({}, 4, None),
            ({i: "test" for i in range(4)}, 4, "test"),
            (
                {i: TendermintRecoveryParams("test") for i in range(4)},
                4,
                TendermintRecoveryParams("test"),
            ),
            ({1: "test", 2: "non-matching", 3: "test", 4: "test"}, 4, "test"),
            ({i: "test" for i in range(4)}, 4, "test"),
            ({1: "no", 2: "result", 3: "matches", 4: ""}, 4, None),
        ),
    )
    def test_get_acn_result(
        self,
        address_to_acn_deliverable: Dict[str, Any],
        n_participants: int,
        expected: Optional[str],
    ) -> None:
        """Test `get_acn_result`."""
        shared_state = SharedState(
            abci_app_cls=AbciAppTest, name="", skill_context=MagicMock()
        )
        shared_state.context.params.setup_params = {
            "test": [],
            "all_participants": ["0x0"],
        }
        shared_state.setup()
        shared_state.synchronized_data.update(participants=tuple(range(n_participants)))
        shared_state.address_to_acn_deliverable = address_to_acn_deliverable
        actual = shared_state.get_acn_result()

        assert actual == expected

    def test_recovery_params_on_init(self) -> None:
        """Test that `tm_recovery_params` get initialized correctly."""
        shared_state = SharedState(name="", skill_context=MagicMock())
        assert shared_state.tm_recovery_params is not None
        assert shared_state.tm_recovery_params.round_count == ROUND_COUNT_DEFAULT
        assert (
            shared_state.tm_recovery_params.reset_from_round
            == AbciAppTest.initial_round_cls.auto_round_id()
        )
        assert shared_state.tm_recovery_params.reset_params is None

    def test_set_last_reset_params(self) -> None:
        """Test that `last_reset_params` get set correctly."""
        shared_state = SharedState(name="", skill_context=MagicMock())
        test_params = [("genesis_time", "some-time"), ("initial_height", "0")]
        shared_state.last_reset_params = test_params
        assert shared_state.last_reset_params == test_params


class TestBenchmarkTool:
    """Test BenchmarkTool"""

    @staticmethod
    def _check_behaviour_data(data: List, agent_name: str) -> None:
        """Check behaviour data."""
        assert len(data) == 1

        (behaviour_data,) = data
        assert behaviour_data["behaviour"] == agent_name
        assert all(
            [key in behaviour_data["data"] for key in ("local", "consensus", "total")]
        )

    def test_end_2_end(self) -> None:
        """Test end 2 end of the tool."""

        agent_name = "agent"
        skill_context = MagicMock(
            agent_address=agent_name, logger=MagicMock(info=logging.info)
        )

        with TemporaryDirectory() as temp_dir:
            benchmark = BenchmarkTool(
                name=agent_name, skill_context=skill_context, log_dir=temp_dir
            )

            with benchmark.measure(agent_name).local():
                sleep(1.0)

            with benchmark.measure(agent_name).consensus():
                sleep(1.0)

            self._check_behaviour_data(benchmark.data, agent_name)

            benchmark.save()

            benchmark_dir = Path(temp_dir, agent_name)
            benchmark_file = benchmark_dir / "0.json"
            assert (benchmark_file).is_file()

            behaviour_data = json.loads(benchmark_file.read_text())
            self._check_behaviour_data(behaviour_data, agent_name)


def test_requests_model_initialization() -> None:
    """Test initialization of the 'Requests(Model)' class."""
    Requests(name="", skill_context=MagicMock())


def test_base_params_model_initialization() -> None:
    """Test initialization of the 'BaseParams(Model)' class."""
    kwargs = BASE_DUMMY_PARAMS.copy()
    bp = BaseParams(**kwargs)

    with pytest.raises(AttributeError, match="This object is frozen!"):
        bp.request_timeout = 0.1

    with pytest.raises(AttributeError, match="This object is frozen!"):
        del bp.request_timeout

    bp.__dict__["_frozen"] = False
    del bp.request_timeout

    assert getattr(bp, "request_timeout", None) is None

    kwargs["skill_context"] = MagicMock(is_abstract_component=False)
    required_setup_params = {
        "safe_contract_address": "0x0",
        "all_participants": ["0x0"],
        "consensus_threshold": 1,
    }
    kwargs["setup"] = required_setup_params
    BaseParams(**kwargs)


@pytest.mark.parametrize(
    "setup, error_text",
    (
        ({}, "`setup` params contain no values!"),
        (
            {"a": "b"},
            "Value for `safe_contract_address` missing from the `setup` params.",
        ),
    ),
)
def test_incorrect_setup(setup: Dict[str, Any], error_text: str) -> None:
    """Test BaseParams model initialization with incorrect setup data."""
    kwargs = BASE_DUMMY_PARAMS.copy()

    with pytest.raises(
        AEAEnforceError,
        match=error_text,
    ):
        kwargs["skill_context"] = MagicMock(is_abstract_component=False)
        kwargs["setup"] = setup
        BaseParams(**kwargs)

    with pytest.raises(
        AEAEnforceError,
        match=f"`reset_pause_duration` must be greater than or equal to {MIN_RESET_PAUSE_DURATION}",
    ):
        kwargs["reset_pause_duration"] = MIN_RESET_PAUSE_DURATION - 1
        BaseParams(**kwargs)


def test_genesis_block() -> None:
    """Test genesis block methods."""
    json = {"max_bytes": "a", "max_gas": "b", "time_iota_ms": "c"}
    gb = GenesisBlock(**json)
    assert gb.to_json() == json

    with pytest.raises(TypeError, match="Error in field 'max_bytes'. Expected type .*"):
        json["max_bytes"] = 0  # type: ignore
        GenesisBlock(**json)


def test_genesis_evidence() -> None:
    """Test genesis evidence methods."""
    json = {"max_age_num_blocks": "a", "max_age_duration": "b", "max_bytes": "c"}
    ge = GenesisEvidence(**json)
    assert ge.to_json() == json


def test_genesis_validator() -> None:
    """Test genesis validator methods."""
    json = {"pub_key_types": ["a", "b"]}
    ge = GenesisValidator(pub_key_types=tuple(json["pub_key_types"]))
    assert ge.to_json() == json

    with pytest.raises(
        TypeError, match="Error in field 'pub_key_types'. Expected type .*"
    ):
        GenesisValidator(**json)  # type: ignore


def test_genesis_consensus_params() -> None:
    """Test genesis consensus params methods."""
    consensus_params = cast(Dict, irrelevant_genesis_config["consensus_params"])
    gcp = GenesisConsensusParams.from_json_dict(consensus_params)
    assert gcp.to_json() == consensus_params


def test_genesis_config() -> None:
    """Test genesis config methods."""
    gcp = GenesisConfig.from_json_dict(irrelevant_genesis_config)
    assert gcp.to_json() == irrelevant_genesis_config


def test_meta_shared_state_when_instance_not_subclass_of_shared_state() -> None:
    """Test instantiation of meta class when instance not a subclass of shared state."""

    class MySharedState(metaclass=_MetaSharedState):
        pass


def test_shared_state_instantiation_without_attributes_raises_error() -> None:
    """Test that definition of concrete subclass of SharedState without attributes raises error."""
    with pytest.raises(AttributeError, match="'abci_app_cls' not set on .*"):

        class MySharedState(BaseSharedState):
            pass

    with pytest.raises(AttributeError, match="The object `None` is not a class"):

        class MySharedStateB(BaseSharedState):
            abci_app_cls = None  # type: ignore

    with pytest.raises(
        AttributeError,
        match="The class <class 'unittest.mock.MagicMock'> is not an instance of packages.valory.skills.abstract_round_abci.base.AbciApp",
    ):

        class MySharedStateC(BaseSharedState):
            abci_app_cls = MagicMock


@dataclass
class A:
    """Class for testing."""

    value: int


@dataclass
class B:
    """Class for testing."""

    value: str


class C(TypedDict):
    """Class for testing."""

    name: str
    year: int


class D(TypedDict, total=False):
    """Class for testing."""

    name: str
    year: int


testdata_positive = [
    ("test_arg", 1, int),
    ("test_arg", "1", str),
    ("test_arg", True, bool),
    ("test_arg", 1, Optional[int]),
    ("test_arg", None, Optional[int]),
    ("test_arg", "1", Optional[str]),
    ("test_arg", None, Optional[str]),
    ("test_arg", None, Optional[bool]),
    ("test_arg", None, Optional[List[int]]),
    ("test_arg", [], Optional[List[int]]),
    ("test_arg", [1], Optional[List[int]]),
    ("test_arg", {"str": 1}, Optional[Dict[str, int]]),
    ("test_arg", {"str": A(1)}, Dict[str, A]),
    ("test_arg", [("1", "2")], List[Tuple[str, str]]),
    ("test_arg", [1], List[Optional[int]]),
    ("test_arg", [1, None], List[Optional[int]]),
    ("test_arg", A, Type[A]),
    ("test_arg", A, Optional[Type[A]]),
    ("test_arg", None, Optional[Type[A]]),
    ("test_arg", MagicMock(), Optional[Type[A]]),  # any type allowed
    ("test_arg", {"name": "str", "year": 1}, C),
    ("test_arg", 42, Literal[42]),
    ("test_arg", {"name": "str"}, D),
]


@pytest.mark.parametrize("name,value,type_hint", testdata_positive)
def test_type_check_positive(name: str, value: Any, type_hint: Any) -> None:
    """Test the type check mixin."""

    check_type(name, value, type_hint)


testdata_negative = [
    ("test_arg", "1", int),
    ("test_arg", 1, str),
    ("test_arg", None, bool),
    ("test_arg", "1", Optional[int]),
    ("test_arg", 1, Optional[str]),
    ("test_arg", 1, Optional[bool]),
    ("test_arg", ["1"], Optional[List[int]]),
    ("test_arg", {"str": "1"}, Optional[Dict[str, int]]),
    ("test_arg", {1: 1}, Optional[Dict[str, int]]),
    ("test_arg", {"str": B("1")}, Dict[str, A]),
    ("test_arg", [()], List[Tuple[str, str]]),
    ("test_arg", [("1",)], List[Tuple[str, str]]),
    ("test_arg", [("1", 1)], List[Tuple[str, str]]),
    ("test_arg", [("1", 1, "1")], List[Tuple[str, ...]]),
    ("test_arg", ["1"], List[Optional[int]]),
    ("test_arg", [1, None, "1"], List[Optional[int]]),
    ("test_arg", B, Type[A]),
    ("test_arg", B, Optional[Type[A]]),
    ("test_arg", {"name": "str", "year": "1"}, C),
    ("test_arg", 41, Literal[42]),
    ("test_arg", C({"name": "str", "year": 1}), A),
    ("test_arg", {"name": "str"}, C),
]


@pytest.mark.parametrize("name,value,type_hint", testdata_negative)
def test_type_check_negative(name: str, value: Any, type_hint: Any) -> None:
    """Test the type check mixin."""

    with pytest.raises(TypeError):
        check_type(name, value, type_hint)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Push all available packages to a registry."""


from pathlib import Path
from typing import Optional

import click
from aea.cli.push_all import push_all_packages
from aea.cli.utils.click_utils import registry_flag

from autonomy.configurations.base import PACKAGE_TYPE_TO_CONFIG_CLASS


@click.command("push-all")
@click.option(
    "--packages-dir", type=click.Path(file_okay=False, dir_okay=True, exists=True)
)
@click.option(
    "--retries",
    type=int,
    default=1,
    help="Tries on package push to the network.",
)
@registry_flag()
def push_all(packages_dir: Optional[Path], retries: int, registry: str) -> None:
    """Push all available packages to a registry."""
    push_all_packages(
        registry,
        packages_dir,
        package_type_config_class=PACKAGE_TYPE_TO_CONFIG_CLASS,
        retries=retries,
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""HTTP server to control the tendermint execution environment."""
import json
import logging
import os
import re
import shutil
import stat
import traceback
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, cast

import requests
from flask import Flask, Response, jsonify, request
from werkzeug.exceptions import InternalServerError, NotFound


try:
    from .tendermint import TendermintNode, TendermintParams  # type: ignore
except ImportError:
    from tendermint import TendermintNode, TendermintParams  # type: ignore

ENCODING = "utf-8"
DEFAULT_LOG_FILE = "log.log"
IS_DEV_MODE = os.environ.get("DEV_MODE", "0") == "1"
CONFIG_OVERRIDE = [
    ("fast_sync = true", "fast_sync = false"),
    ("max_num_outbound_peers = 10", "max_num_outbound_peers = 0"),
    ("pex = true", "pex = false"),
]
DOCKER_INTERNAL_HOST = "host.docker.internal"
TM_STATUS_ENDPOINT = "http://localhost:26657/status"

logging.basicConfig(
    filename=os.environ.get("LOG_FILE", DEFAULT_LOG_FILE),
    level=logging.DEBUG,
    format="%(asctime)s %(levelname)s %(name)s %(threadName)s : %(message)s",  # noqa : W1309
)


def load_genesis() -> Any:
    """Load genesis file."""
    return json.loads(
        Path(os.environ["TMHOME"], "config", "genesis.json").read_text(
            encoding=ENCODING
        )
    )


def get_defaults() -> Dict[str, str]:
    """Get defaults from genesis file."""
    genesis = load_genesis()
    return dict(genesis_time=genesis.get("genesis_time"))


def override_config_toml() -> None:
    """Update sync method."""

    config_path = str(Path(os.environ["TMHOME"]) / "config" / "config.toml")
    logging.info(config_path)
    with open(config_path, "r", encoding=ENCODING) as fp:
        config = fp.read()

    for old, new in CONFIG_OVERRIDE:
        config = config.replace(old, new)

    with open(config_path, "w+", encoding=ENCODING) as fp:
        fp.write(config)


def update_peers(validators: List[Dict], config_path: Path) -> None:
    """Fix peers."""

    config_text = config_path.read_text(encoding="utf-8")

    new_peer_string = 'persistent_peers = "'
    for peer in validators:
        hostname = peer["hostname"]
        if hostname in ("localhost", "0.0.0.0"):
            # This (tendermint) node will be running in a docker container and no other node
            # will be running in the same container. If we receive either localhost or 0.0.0.0,
            # we make an assumption that the address belongs to a node running on the
            # same machine with a different docker container and different p2p port so,
            # we replace the hostname with the docker's internal host url.
            hostname = DOCKER_INTERNAL_HOST
        new_peer_string += (
            peer["peer_id"] + "@" + hostname + ":" + str(peer["p2p_port"]) + ","
        )
    new_peer_string = new_peer_string[:-1] + '"\n'

    updated_config = re.sub('persistent_peers = ".*\n', new_peer_string, config_text)
    config_path.write_text(updated_config, encoding="utf-8")


def update_external_address(external_address: str, config_path: Path) -> None:
    """Update the external address."""
    config_text = config_path.read_text(encoding="utf-8")
    new_external_address = f'external_address = "{external_address}"\n'
    updated_config = re.sub(
        'external_address = ".*\n', new_external_address, config_text
    )
    config_path.write_text(updated_config, encoding="utf-8")


def update_genesis_config(data: Dict) -> None:
    """Update genesis.json file for the tendermint node."""

    genesis_file = Path(os.environ["TMHOME"]) / "config" / "genesis.json"
    genesis_data = {}
    genesis_data["genesis_time"] = data["genesis_config"]["genesis_time"]
    genesis_data["chain_id"] = data["genesis_config"]["chain_id"]
    genesis_data["initial_height"] = "0"
    genesis_data["consensus_params"] = data["genesis_config"]["consensus_params"]
    genesis_data["validators"] = [
        {
            "address": validator["address"],
            "pub_key": validator["pub_key"],
            "power": validator["power"],
            "name": validator["name"],
        }
        for validator in data["validators"]
    ]
    genesis_data["app_hash"] = ""
    genesis_file.write_text(json.dumps(genesis_data, indent=2), encoding=ENCODING)


class PeriodDumper:
    """Dumper for tendermint data."""

    resets: int
    dump_dir: Path
    logger: logging.Logger

    def __init__(self, logger: logging.Logger, dump_dir: Optional[Path] = None) -> None:
        """Initialize object."""

        self.resets = 0
        self.logger = logger
        self.dump_dir = dump_dir or Path("/tm_state")

        if self.dump_dir.is_dir():
            shutil.rmtree(str(self.dump_dir), onerror=self.readonly_handler)
        self.dump_dir.mkdir(exist_ok=True)

    @staticmethod
    def readonly_handler(
        func: Callable, path: str, execinfo: Any  # pylint: disable=unused-argument
    ) -> None:
        """If permission is readonly, we change and retry."""
        try:
            os.chmod(path, stat.S_IWRITE)
            func(path)
        except (FileNotFoundError, OSError):
            return

    def dump_period(self) -> None:
        """Dump tendermint run data for replay"""
        store_dir = self.dump_dir / f"period_{self.resets}"
        store_dir.mkdir(exist_ok=True)
        try:
            shutil.copytree(
                os.environ["TMHOME"], str(store_dir / ("node" + os.environ["ID"]))
            )
            self.logger.info(f"Dumped data for period {self.resets}")
        except OSError as e:
            self.logger.info(
                f"Error occurred while dumping data for period {self.resets}: {e}"
            )
        self.resets += 1


def create_app(  # pylint: disable=too-many-statements
    dump_dir: Optional[Path] = None,
    perform_monitoring: bool = True,
    debug: bool = False,
) -> Tuple[Flask, TendermintNode]:
    """Create the Tendermint server app"""

    tendermint_params = TendermintParams(
        proxy_app=os.environ["PROXY_APP"],
        consensus_create_empty_blocks=os.environ["CREATE_EMPTY_BLOCKS"] == "true",
        home=os.environ["TMHOME"],
        use_grpc=os.environ["USE_GRPC"] == "true",
    )

    app = Flask(__name__)
    period_dumper = PeriodDumper(logger=app.logger, dump_dir=dump_dir)
    tendermint_node = TendermintNode(tendermint_params, logger=app.logger)
    tendermint_node.init()

    override_config_toml()

    tendermint_node.start(start_monitoring=perform_monitoring, debug=debug)

    @app.get("/params")
    def get_params() -> Dict:
        """Get tendermint params."""
        try:
            priv_key_file = (
                Path(os.environ["TMHOME"]) / "config" / "priv_validator_key.json"
            )
            priv_key_data = json.loads(priv_key_file.read_text(encoding=ENCODING))
            del priv_key_data["priv_key"]
            status = requests.get(TM_STATUS_ENDPOINT).json()
            priv_key_data["peer_id"] = status["result"]["node_info"]["id"]
            return {
                "params": priv_key_data,
                "status": True,
                "error": None,
            }
        except (FileNotFoundError, json.JSONDecodeError):
            return {"params": {}, "status": False, "error": traceback.format_exc()}

    @app.post("/params")
    def update_params() -> Dict:
        """Update validator params."""

        try:
            data: Dict = json.loads(request.get_data().decode(ENCODING))
            cast(logging.Logger, app.logger).debug(  # pylint: disable=no-member
                f"Data update requested with data={data}"
            )

            cast(logging.Logger, app.logger).info(  # pylint: disable=no-member
                "Updating genesis config."
            )
            update_genesis_config(data=data)

            cast(logging.Logger, app.logger).info(  # pylint: disable=no-member
                "Updating peristent peers."
            )
            config_path = Path(os.environ["TMHOME"]) / "config" / "config.toml"
            update_peers(
                validators=data["validators"],
                config_path=config_path,
            )
            update_external_address(
                external_address=data["external_address"],
                config_path=config_path,
            )

            return {"status": True, "error": None}
        except (FileNotFoundError, json.JSONDecodeError, PermissionError):
            return {"status": False, "error": traceback.format_exc()}

    @app.route("/gentle_reset")
    def gentle_reset() -> Tuple[Any, int]:
        """Reset the tendermint node gently."""
        try:
            tendermint_node.stop()
            tendermint_node.start(start_monitoring=perform_monitoring)
            return jsonify({"message": "Reset successful.", "status": True}), 200
        except Exception as e:  # pylint: disable=W0703
            return jsonify({"message": f"Reset failed: {e}", "status": False}), 200

    @app.route("/app_hash")
    def app_hash() -> Tuple[Any, int]:
        """Get the app hash."""
        try:
            non_routable, loopback = "0.0.0.0", "127.0.0.1"
            endpoint = f"{tendermint_params.rpc_laddr.replace('tcp', 'http').replace(non_routable, loopback)}/block"
            height = request.args.get("height")
            params = {"height": height} if height is not None else None
            res = requests.get(endpoint, params)
            app_hash_ = res.json()["result"]["block"]["header"]["app_hash"]
            return jsonify({"app_hash": app_hash_}), res.status_code
        except Exception as e:  # pylint: disable=W0703
            return (
                jsonify({"error": f"Could not get the app hash: {str(e)}"}),
                200,
            )

    @app.route("/hard_reset")
    def hard_reset() -> Tuple[Any, int]:
        """Reset the node forcefully, and prune the blocks"""
        try:
            tendermint_node.stop()
            if IS_DEV_MODE:
                period_dumper.dump_period()

            return_code = tendermint_node.prune_blocks()
            if return_code:
                tendermint_node.start(start_monitoring=perform_monitoring)
                raise RuntimeError("Could not perform `unsafe-reset-all` successfully!")
            defaults = get_defaults()
            tendermint_node.reset_genesis_file(
                request.args.get("genesis_time", defaults["genesis_time"]),
                # default should be 1: https://github.com/tendermint/tendermint/pull/5191/files
                request.args.get("initial_height", "1"),
                request.args.get("period_count", "0"),
            )
            tendermint_node.start(start_monitoring=perform_monitoring)
            return jsonify({"message": "Reset successful.", "status": True}), 200
        except Exception as e:  # pylint: disable=W0703
            return jsonify({"message": f"Reset failed: {e}", "status": False}), 200

    @app.errorhandler(404)  # type: ignore
    def handle_notfound(e: NotFound) -> Response:
        """Handle server error."""
        cast(logging.Logger, app.logger).info(e)  # pylint: disable=E
        return Response("Not Found", status=404, mimetype="application/json")

    @app.errorhandler(500)  # type: ignore
    def handle_server_error(e: InternalServerError) -> Response:
        """Handle server error."""
        cast(logging.Logger, app.logger).info(e)  # pylint: disable=E
        return Response("Error Closing Node", status=500, mimetype="application/json")

    return app, tendermint_node


def create_server() -> Any:
    """Function to retrieve just the app to be used by flask entry point."""
    flask_app, _ = create_app()
    return flask_app



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the code-blocks in the documentation."""

from pathlib import Path
from typing import Callable, Dict, List, Optional, cast

from tests.conftest import ROOT_DIR
from tests.test_docs.helper import (  # type: ignore
    CodeType,
    check_bash_commands_exist,
    check_code_blocks_exist,
    contains_code_blocks,
    extract_make_commands,
    remove_doc_ellipsis,
    remove_line_comments,
)


class BaseTestDocCode:
    """Base class for doc code testing"""

    md_to_code: Dict[str, Dict] = {}
    code_type: CodeType = CodeType.NOCODE
    skipped_files: Optional[List[str]] = None

    def _to_os_path(self, file_path: str) -> str:
        r"""
        Transforms a file path to a path in the OS the code is being executed on.

        Example:
        Given file_path="docs/fsm.yaml",
        this method would transform it to:
         - "docs/fsm.yaml" for POSIX systems. (no changes)
         - "docs\\fsm.yaml" for Windows systems.

        :param file_path: the file path to transform.
        :return: the transformed file path
        """
        path = Path(file_path)
        return str(path)

    def test_run_check(self) -> None:
        """Check the documentation code"""

        assert (
            self.code_type != CodeType.NOCODE
        ), "This test class has not specified a code type to check for"
        assert (
            self.md_to_code != {}
        ), "This test class has not specified the md_to_code mapping"

        # Get all doc files that contain a block
        all_md_files = [
            p.relative_to(ROOT_DIR) for p in Path(ROOT_DIR, "docs").rglob("*.md")
        ]
        files_with_blocks = list(
            map(
                str,
                filter(
                    cast(
                        Callable[[Path], bool],
                        lambda f: "api" not in f.parts  # skip api folder
                        and contains_code_blocks(f, self.code_type.value),
                    ),
                    all_md_files,
                ),
            )
        )

        if self.skipped_files:
            os_path_skipped_files = {
                self._to_os_path(file) for file in self.skipped_files
            }
            files_with_blocks = [
                f for f in files_with_blocks if f not in os_path_skipped_files
            ]
        os_path_md_to_code = {self._to_os_path(file) for file in self.md_to_code.keys()}
        not_checked_files = set(files_with_blocks).difference(os_path_md_to_code)

        assert (
            not not_checked_files
        ), f"The following (not skipped) doc files contain {self.code_type.value} blocks but are not being checked: {not_checked_files}"

        # Check all files
        for md_file, code_info in self.md_to_code.items():
            print(
                f"Checking {self.code_type.value} snippets in file {md_file}... ",
                end="",
            )

            # Preprocessing functions:
            # - For doc files: `doc_process_fn` -> remove tokens like "# ...\n" from the code
            # - For code files: code_process_fn` not required.
            check_code_blocks_exist(
                md_file=md_file,
                code_info=code_info,
                code_type=self.code_type,
                doc_process_fn=lambda s: remove_doc_ellipsis(remove_line_comments(s)),
            )

            print("OK")


class TestYamlSnippets(BaseTestDocCode):
    """Test that all the yaml snippets in the documentation exist in the repository"""

    code_type = CodeType.YAML

    # This variable holds a mapping between every doc file and the code files
    # that contains the referenced code. Since a doc file can contain several code
    # snippets, a list with the target files ordered is provided.
    #
    # Use skip_blocks to specify a list of blocks that need to be skipped or add a file to skipped_files
    # to skip it completely.
    # Add by_line:: at the beginning of a code file path so the check is performed line by line
    # instead of checking the code block as a whole.

    md_to_code = {
        # "docs/demos/hello_world_demo.md": {
        #     "code_files": [
        #         "packages/valory/skills/hello_world_abci/fsm_specification.yaml",
        #         "packages/valory/agents/hello_world/aea-config.yaml",
        #     ],
        # },
        "docs/demos/price_oracle_fsms.md": {
            "code_files": [
                "packages/valory/skills/registration_abci/fsm_specification.yaml",
            ],
            "skip_blocks": [1, 2, 3, 4, 5],
        },
        "docs/guides/deploy_service.md": {"skip_blocks": [0]},
        "docs/advanced_reference/developer_tooling/benchmarking.md": {
            "skip_blocks": [0]
        },
        "docs/guides/draft_service_idea_and_define_fsm_specification.md": {
            "code_files": [
                "packages/valory/skills/hello_world_abci/fsm_specification.yaml"
            ]
        },
    }

    skipped_files = [
        "docs/guides/define_agent.md",  # TODO: How to check against hello w.? only changes name of vendor and agent.
        "docs/guides/define_service.md",  # TODO: How to check against hello w.? only changes name of vendor and service
        "docs/configure_service/service_configuration_file.md",
        "docs/configure_service/on-chain_deployment_checklist.md",  # just placeholder examples
        "docs/configure_service/configure_access_external_chains.md",  # just placeholder examples
        "docs/advanced_reference/developer_tooling/dev_mode.md",  # just placeholder examples
        "docs/demos/hello_world_demo.md",  # TODO: Fix the hash update script to update the aea config code snippet
    ]


class TestPythonSnippets(BaseTestDocCode):
    """Test that all the python snippets in the documentation exist in the repository"""

    code_type = CodeType.PYTHON

    # This variable holds a mapping between every doc file and the code file
    # that contains the referenced code. Since a doc file can contain several code
    # snippets, a list with the target files ordered is provided.
    #
    # Use skip_blocks to specify a list of blocks that need to be skipped
    # Add by_line:: at the beginning of a code file path so the check is performed line by line
    # instead of checking the code block as a whole.

    md_to_code = {
        "docs/key_concepts/abci_app_abstract_round_behaviour.md": {
            "code_files": [
                "by_line::packages/valory/skills/abstract_round_abci/behaviours.py"
            ],
            "skip_blocks": [1],
        },
        "docs/key_concepts/abci_app_class.md": {
            "code_files": [
                "by_line::packages/valory/skills/abstract_round_abci/base.py"
            ],
            "skip_blocks": [1],
        },
        "docs/demos/hello_world_demo.md": {
            "code_files": [
                "packages/valory/skills/hello_world_abci/behaviours.py",
                "packages/valory/skills/hello_world_abci/behaviours.py",
                "packages/valory/skills/hello_world_abci/payloads.py",
                "packages/valory/skills/hello_world_abci/rounds.py",
                "by_line::packages/valory/skills/hello_world_abci/rounds.py",
            ],
        },
        "docs/advanced_reference/commands/autonomy_analyse.md": {"skip_blocks": [0]},
        "docs/key_concepts/aea.md": {"code_files": [], "skip_blocks": [0, 1]},
    }

    skipped_files = [
        "docs/key_concepts/abci_app_async_behaviour.md",  # just placeholder examples
        "docs/configure_service/configure_access_external_chains.md",  # only irrelevant one-liners,
        "docs/key_concepts/abci_app_abstract_round.md",  # just a placeholder example
        "docs/demos/price_oracle_fsms.md",  # price oracle has been extracted to a separate repo on #1441
        "docs/demos/price_oracle_technical_details.md",  # price oracle has been extracted to a separate repo on #1441
        "docs/advanced_reference/developer_tooling/benchmarking.md",  # just placeholder examples
        "docs/configure_service/on-chain_deployment_checklist.md",  # just placeholder examples
    ]


class TestJsonSnippets(BaseTestDocCode):
    """Test that all the yaml snippets in the documentation exist in the repository"""

    code_type = CodeType.JSON

    # This variable holds a mapping between every doc file and the code files
    # that contains the referenced code. Since a doc file can contain several code
    # snippets, a list with the target files ordered is provided.
    #
    # Use skip_blocks to specify a list of blocks that need to be skipped or add a file to skipped_files
    # to skip it completely.
    # Add by_line:: at the beginning of a code file path so the check is performed line by line
    # instead of checking the code block as a whole.

    md_to_code = {
        "docs/guides/set_up.md": {
            "code_files": ["by_line::packages/packages.json"],
        },
        "docs/guides/deploy_service.md": {
            "code_files": [
                "by_line::deployments/keys/hardhat_keys.json",
                "by_line::deployments/keys/hardhat_keys.json",
            ],
        },
        "docs/guides/quick_start.md": {
            "code_files": ["by_line::deployments/keys/hardhat_keys.json"],
        },
        "docs/advanced_reference/developer_tooling/dev_mode.md": {
            "code_files": ["by_line::deployments/keys/hardhat_keys.json"]
        },
        "docs/counter_example.md": {
            "code_files": ["by_line::deployments/keys/hardhat_keys.json"]
        },
    }

    skipped_files = [
        "docs/advanced_reference/commands/autonomy_deploy.md",
    ]


class TestDocBashSnippets:
    """Class for doc bash snippet testing"""

    def test_run_check(self) -> None:
        """Check the documentation code"""

        code_type = CodeType.BASH

        skipped_files: List[str] = []

        # Get all doc files that contain a block
        all_md_files = [
            str(p.relative_to(ROOT_DIR)) for p in Path(ROOT_DIR, "docs").rglob("*.md")
        ]
        files_with_blocks = list(
            filter(
                lambda f: "/api/" not in f  # skip api folder
                and contains_code_blocks(f, code_type.value),
                all_md_files,
            )
        )
        if skipped_files:
            files_with_blocks = [f for f in files_with_blocks if f not in skipped_files]

        all_mk_files = [str(p.absolute()) for p in Path(ROOT_DIR).rglob("*Makefile")]

        all_mk_files = list(
            filter(
                lambda f: "third_party/" not in f
                and ".tox" not in f,  # skip some folders
                all_mk_files,
            )
        )

        make_commands = extract_make_commands(all_mk_files)

        for md_file in files_with_blocks:
            check_bash_commands_exist(
                md_file,
                make_commands,
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Override for packages command."""

import sys
from pathlib import Path
from typing import List, Tuple, cast
from warnings import warn

import click
from aea.cli.packages import package_manager, package_type_selector_prompt
from aea.cli.utils.click_utils import reraise_as_click_exception
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import pass_ctx
from aea.configurations.base import PackageConfiguration
from aea.configurations.constants import PACKAGE_TYPE_TO_CONFIG_FILE
from aea.configurations.data_types import PackageId, PackageType
from aea.helpers.dependency_tree import dump_yaml, load_yaml
from aea.package_manager.base import (
    BasePackageManager,
    DepedencyMismatchErrors,
    PackageFileNotValid,
)
from aea.package_manager.v0 import PackageManagerV0 as BasePackageManagerV0
from aea.package_manager.v1 import PackageManagerV1 as BasePackageManagerV1

from autonomy.cli.helpers.ipfs_hash import load_configuration
from autonomy.configurations.base import Service


@package_manager.command(name="lock")
@click.option(
    "--check",
    is_flag=True,
    help="Check that fingerprints in packages.json match the local packages",
)
@click.option(
    "--skip-missing",
    is_flag=True,
    help="Skip packages missing from the `packages.json` file.",
)
@pass_ctx
def lock_packages(ctx: Context, check: bool, skip_missing: bool) -> None:
    """Lock local packages."""

    packages_dir = Path(ctx.registry_path)

    with reraise_as_click_exception(Exception):
        if check:
            click.echo("Verifying packages.json")
            return_code = get_package_manager(packages_dir).verify()

            if return_code:
                click.echo("Verification failed.")
            else:
                click.echo("Verification successful")

            sys.exit(return_code)

        click.echo("Updating hashes...")
        get_package_manager(packages_dir).update_package_hashes(
            selector_prompt=package_type_selector_prompt,
            skip_missing=skip_missing,
        ).dump()
        click.echo("Done")


def get_package_manager(packages_dir: Path) -> BasePackageManager:
    """Get package manager."""

    try:
        return PackageManagerV1.from_dir(
            packages_dir=packages_dir, config_loader=load_configuration
        )
    except PackageFileNotValid:
        warn(
            "The provided `packages.json` still follows an older format which will be deprecated on v2.0.0",
            DeprecationWarning,
            stacklevel=2,
        )
        click.echo(
            "The provided `packages.json` still follows an older format which will be deprecated on v2.0.0"
        )
        return PackageManagerV0.from_dir(
            packages_dir=packages_dir, config_loader=load_configuration
        )


class _PackageManagerWithServicePatch(BasePackageManager):
    """Patch package manager for service component."""

    def update_dependencies(self, package_id: PackageId) -> None:
        """Update dependencies."""

        if package_id.package_type != PackageType.SERVICE:  # pragma: nocover
            super().update_dependencies(package_id=package_id)
            return

        package_path = self.package_path_from_package_id(
            package_id=package_id,
        )
        config_file = (
            package_path / PACKAGE_TYPE_TO_CONFIG_FILE[package_id.package_type.value]
        )
        package_config, extra = load_yaml(file_path=config_file)
        package_config["agent"] = self.update_public_id_hash(
            public_id_str=package_config["agent"],
            package_type=PackageType.AGENT,
        )

        dump_yaml(
            file_path=config_file,
            data=package_config,
            extra_data=extra,
        )

    def check_dependencies(
        self, configuration: PackageConfiguration
    ) -> List[Tuple[PackageId, DepedencyMismatchErrors]]:
        """Update dependencies."""

        if configuration.package_type != PackageType.SERVICE:  # pragma: nocover
            return super().check_dependencies(
                configuration=configuration,
            )

        configuration = cast(Service, configuration)
        agent_id = PackageId(
            package_type=PackageType.AGENT,
            public_id=configuration.agent,
        )

        expected_hash = self.get_package_hash(package_id=agent_id)
        if expected_hash is None:
            return [
                (agent_id, DepedencyMismatchErrors.HASH_NOT_FOUND),
            ]

        if expected_hash != agent_id.package_hash:
            return [
                (agent_id, DepedencyMismatchErrors.HASH_DOES_NOT_MATCH),
            ]

        return []


class PackageManagerV0(BasePackageManagerV0, _PackageManagerWithServicePatch):
    """Patch package manager for service component."""


class PackageManagerV1(BasePackageManagerV1, _PackageManagerWithServicePatch):
    """Patch package manager for service component."""



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test for contract module."""

from pathlib import Path
from unittest import mock

from aea_test_autonomy.base_test_classes.contracts import BaseRegistriesContractsTest
from aea_test_autonomy.docker.base import skip_docker_tests
from aea_test_autonomy.docker.registries import AGENT_REGISTRY

from packages.valory.contracts.agent_registry.contract import (
    AGENT_UNIT_TYPE,
    AgentRegistryContract,
    UNIT_HASH_PREFIX,
)


PACKAGE_DIR = Path(__file__).parent.parent


@skip_docker_tests
class TestAgentRegistry(BaseRegistriesContractsTest):
    """Test agent registry."""

    contract: AgentRegistryContract
    contract_address = AGENT_REGISTRY
    contract_directory = PACKAGE_DIR

    def test_get_token_uri(self) -> None:
        """Test get token URI method."""

        token_uri = self.contract.get_token_uri(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            token_id=1,
        )

        assert (
            token_uri
            == "https://gateway.autonolas.tech/ipfs/f01701220985b4c36158b51f8a865faceff8141dbc0989c349a1a41ba1e2ac8e5b24536b2"  # nosec
        )

    def test_filter_token_id_from_emitted_events(self) -> None:
        """Test `filter_token_id_from_emitted_events` method"""

        dummy_hash = b"dummy_hash"
        expected_hash = UNIT_HASH_PREFIX.format(metadata_hash=dummy_hash.hex())

        with mock.patch.object(
            AgentRegistryContract,
            "get_instance",
            return_value=mock.MagicMock(
                events=mock.MagicMock(
                    CreateUnit=mock.MagicMock(
                        createFilter=lambda **_: mock.MagicMock(
                            get_all_entries=lambda *_: []
                        )
                    )
                )
            ),
        ):
            token_id = self.contract.filter_token_id_from_emitted_events(
                ledger_api=self.ledger_api,
                contract_address=self.contract_address,
                metadata_hash=expected_hash,
            )
            assert token_id is None

        with mock.patch.object(
            AgentRegistryContract,
            "get_instance",
            return_value=mock.MagicMock(
                events=mock.MagicMock(
                    CreateUnit=mock.MagicMock(
                        createFilter=lambda **_: mock.MagicMock(
                            get_all_entries=lambda *_: [
                                {
                                    "args": {
                                        "uType": AGENT_UNIT_TYPE,
                                        "unitHash": dummy_hash,
                                        "unitId": 1,
                                    }
                                }
                            ]
                        )
                    )
                )
            ),
        ), mock.patch.object(
            self.ledger_api.api, "toBytes", return_value=expected_hash.encode()
        ):
            token_id = self.contract.filter_token_id_from_emitted_events(
                ledger_api=self.ledger_api,
                contract_address=self.contract_address,
                metadata_hash=expected_hash,
            )
            assert token_id is not None
            assert token_id == 1



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
# flake8: noqa: F811

"""Integration tests for the valory/register_reset skill."""

# pylint: skip-file

from pathlib import Path

import pytest
from aea.configurations.data_types import PublicId
from aea_test_autonomy.base_test_classes.agents import (
    BaseTestEnd2EndExecution,
    RoundChecks,
)
from aea_test_autonomy.configurations import KEY_PAIRS
from aea_test_autonomy.fixture_helpers import (  # noqa: F401
    UseACNNode,
    UseRegistries,
    abci_host,
    abci_port,
    acn_config,
    acn_node,
    flask_tendermint,
    hardhat_port,
    ipfs_daemon,
    ipfs_domain,
    key_pairs,
    nb_nodes,
    registries_scope_class,
    tendermint_port,
)

from packages.valory.agents.register_reset.tests.helpers.conftest import (  # noqa: F401
    flask_tendermint,
)
from packages.valory.skills.registration_abci.rounds import (
    RegistrationRound,
    RegistrationStartupRound,
)
from packages.valory.skills.reset_pause_abci.rounds import ResetAndPauseRound


HAPPY_PATH = (
    RoundChecks(RegistrationStartupRound.auto_round_id()),
    RoundChecks(RegistrationRound.auto_round_id(), n_periods=2),
    RoundChecks(ResetAndPauseRound.auto_round_id(), n_periods=3),
)


@pytest.mark.e2e
@pytest.mark.integration
@pytest.mark.flaky(reruns=1)
@pytest.mark.parametrize("nb_nodes", (4,))
class TestRaceConditionTendermintReset(UseACNNode, BaseTestEnd2EndExecution):
    """Test that ABCI register-reset skill with 4 agents when resetting Tendermint, with a slow Tendermint server."""

    agent_package = "valory/register_reset:0.1.0"
    skill_package = "valory/register_reset_abci:0.1.0"
    happy_path = HAPPY_PATH
    key_pairs = KEY_PAIRS
    wait_to_finish = 200
    __reset_tendermint_every = 1
    package_registry_src_rel = Path(__file__).parents[4]
    __args_prefix = f"vendor.valory.skills.{PublicId.from_str(skill_package).name}.models.params.args"
    extra_configs = [
        {
            "dotted_path": f"{__args_prefix}.reset_tendermint_after",
            "value": __reset_tendermint_every,
        },
        {
            "dotted_path": f"{__args_prefix}.reset_pause_duration",
            "value": 15,
        },
    ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Script for generating deployment environments."""
from pathlib import Path
from typing import Dict, List, Optional, Type

from autonomy.deploy.base import BaseDeploymentGenerator, ServiceBuilder
from autonomy.deploy.constants import DEPLOYMENT_REPORT, INFO
from autonomy.deploy.generators.docker_compose.base import DockerComposeGenerator
from autonomy.deploy.generators.kubernetes.base import KubernetesGenerator


DEPLOYMENT_OPTIONS: Dict[str, Type[BaseDeploymentGenerator]] = {
    "kubernetes": KubernetesGenerator,
    "docker-compose": DockerComposeGenerator,
}


def generate_deployment(  # pylint: disable=too-many-arguments, too-many-locals
    type_of_deployment: str,
    keys_file: Path,
    service_path: Path,
    build_dir: Path,
    number_of_agents: Optional[int] = None,
    private_keys_password: Optional[str] = None,
    dev_mode: bool = False,
    packages_dir: Optional[Path] = None,
    open_aea_dir: Optional[Path] = None,
    open_autonomy_dir: Optional[Path] = None,
    agent_instances: Optional[List[str]] = None,
    multisig_address: Optional[str] = None,
    consensus_threshold: Optional[int] = None,
    log_level: str = INFO,
    apply_environment_variables: bool = False,
    image_version: Optional[str] = None,
    use_hardhat: bool = False,
    use_acn: bool = False,
    use_tm_testnet_setup: bool = False,
    image_author: Optional[str] = None,
) -> str:
    """Generate the deployment for the service."""

    service_builder = ServiceBuilder.from_dir(
        path=service_path,
        keys_file=keys_file,
        number_of_agents=number_of_agents,
        private_keys_password=private_keys_password,
        agent_instances=agent_instances,
        apply_environment_variables=apply_environment_variables,
    )
    service_builder.deplopyment_type = type_of_deployment
    service_builder.log_level = log_level
    service_builder.try_update_runtime_params(
        multisig_address=multisig_address,
        agent_instances=agent_instances,
        consensus_threshold=consensus_threshold,
    )
    service_builder.try_update_abci_connection_params()

    DeploymentGenerator = DEPLOYMENT_OPTIONS[type_of_deployment]
    deployment = DeploymentGenerator(
        service_builder=service_builder,
        build_dir=build_dir,
        dev_mode=dev_mode,
        packages_dir=packages_dir,
        open_aea_dir=open_aea_dir,
        open_autonomy_dir=open_autonomy_dir,
        use_tm_testnet_setup=use_tm_testnet_setup,
        image_author=image_author,
    )

    (
        deployment.generate(
            image_version=image_version,
            use_acn=use_acn,
            use_hardhat=use_hardhat,
        )
        .generate_config_tendermint()
        .write_config()
        .populate_private_keys()
    )

    return DEPLOYMENT_REPORT.substitute(
        **{
            "type": type_of_deployment,
            "agents": service_builder.service.number_of_agents,
            "size": len(deployment.output),
        }
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test 'scaffold fsm' subcommand."""

import importlib.util
import json
import os
import shutil
from contextlib import suppress
from copy import copy
from importlib.machinery import ModuleSpec
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List, Optional

import click.testing
import pytest
from aea.cli.utils.config import get_default_author_from_cli_config
from aea.package_manager.base import PACKAGES_FILE
from aea.test_tools.test_cases import AEATestCaseMany

# trigger population of autonomy commands
import autonomy.cli.core  # noqa

from packages.valory import skills
from packages.valory.skills.abstract_round_abci.base import _MetaPayload


VALORY_SKILLS_PATH = Path(os.path.join(*skills.__package__.split("."))).absolute()
fsm_specifications = list(VALORY_SKILLS_PATH.glob("**/fsm_specification.yaml"))


class BaseScaffoldFSMTest(AEATestCaseMany):
    """Test `scaffold fsm` subcommand."""

    _t: TemporaryDirectory
    agent_name: str = "test_agent"

    old_tx_type_to_payload_cls: Dict

    @classmethod
    def setup_class(cls) -> None:
        """Setup class."""
        super().setup_class()

        cls.author = get_default_author_from_cli_config()

    def setup(
        self,
    ) -> None:
        """Setup test."""

        # we need to store the current value of the meta-class attribute
        # _MetaPayload.transaction_type_to_payload_cls, and restore it
        # in the teardown function. We do a shallow copy so we avoid
        # to modify the old mapping during the execution of the tests.
        self.old_tx_type_to_payload_cls = copy(_MetaPayload.registry)
        _MetaPayload.registry = {}

        self.run_cli_command(
            "create",
            "--local",
            "--empty",
            self.agent_name,
            "--author",
            self.author,
        )

    def teardown(
        self,
    ) -> None:
        """Teardown test."""
        _MetaPayload.registry = self.old_tx_type_to_payload_cls  # type: ignore
        with suppress(OSError, FileExistsError, PermissionError):
            shutil.rmtree(str(Path(self.t, self.agent_name)))

    def scaffold_fsm(
        self,
        fsm_spec_file: Path,
        scaffold_args: Optional[List] = None,
        cli_args: Optional[List] = None,
    ) -> click.testing.Result:
        """Scaffold FSM."""
        scaffold_args = scaffold_args or []
        cli_args = cli_args or []
        *_, skill_name, _ = fsm_spec_file.parts
        skill_name = f"test_skill_{skill_name}"
        scaffold_args = [
            *cli_args,
            "scaffold",
            *scaffold_args,
            "fsm",
            skill_name,
            "--local",
            "--spec",
        ]
        cli_args = [
            *scaffold_args,
            fsm_spec_file,
        ]
        scaffold_result = self.run_cli_command(*cli_args, cwd=self.t / self.agent_name)

        return scaffold_result


class TestScaffoldFSM(BaseScaffoldFSMTest):
    """Test `scaffold fsm` subcommand."""

    @pytest.mark.parametrize("fsm_spec_file", fsm_specifications)
    def test_scaffold_fsm(
        self, fsm_spec_file: Path, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Test scaffold files are loadable."""

        *_, skill_name, _ = fsm_spec_file.parts
        skill_name = f"test_skill_{skill_name}"
        scaffold_result = self.scaffold_fsm(fsm_spec_file)
        assert scaffold_result.exit_code == 0

        monkeypatch.syspath_prepend(self.t)
        path = self.t / self.author / "skills" / skill_name
        for file in path.rglob("**/*.py"):
            module_spec = importlib.util.spec_from_file_location("name", file)
            assert isinstance(module_spec, ModuleSpec)
            module_type = importlib.util.module_from_spec(module_spec)
            module_spec.loader.exec_module(module_type)  # type: ignore

        # check spec file is in skill directory
        assert (
            self.t / self.agent_name / "skills" / skill_name / fsm_spec_file.name
        ).exists(), "spec file not copied in scaffolded skill"


class TestScaffoldFSMAutonomyTests(BaseScaffoldFSMTest):
    """Test `scaffold fsm` subcommand."""

    @pytest.mark.parametrize("fsm_spec_file", fsm_specifications)
    def test_autonomy_test(self, fsm_spec_file: Path) -> None:
        """Run autonomy test on the scaffolded skill"""

        *_, skill_name, _ = fsm_spec_file.parts
        skill_name = f"test_skill_{skill_name}"
        scaffold_result = self.scaffold_fsm(fsm_spec_file)
        assert scaffold_result.exit_code == 0
        cli_args = [
            "-m",
            "aea.cli",
            "test",
            "by-path",
            str(self.t / self.agent_name / "skills" / skill_name),
        ]

        # we use a subprocess rather than click.CliRunner because we want to isolate
        # the Python environment of the current pytest process with the subcall to pytest.main
        # from the AEA command `aea test by-path ...`
        result = self.start_subprocess(*cli_args)
        result.wait(timeout=60.0)
        assert (
            result.returncode == 0
        ), f"stdout: {self.stdout[result.pid]}, stderr: {self.stderr[result.pid]}"


class TestScaffoldFSMLocalRegistry(BaseScaffoldFSMTest):
    """Test `scaffold fsm -tlr` subcommand."""

    @property
    def packages_path(self) -> Path:
        """
        Get packages path.

        :return: Path
        """
        return self.t / "packages"

    @pytest.mark.parametrize("fsm_spec_file", fsm_specifications)
    def test_autonomy_test(self, fsm_spec_file: Path) -> None:
        """Run autonomy test on the scaffolded skill"""

        *_, skill_name, _ = fsm_spec_file.parts
        skill_name = f"test_skill_{skill_name}"
        scaffold_result = self.scaffold_fsm(
            fsm_spec_file,
            scaffold_args=["-tlr"],
            cli_args=["--registry-path", self.packages_path],
        )
        assert scaffold_result.exit_code == 0
        assert (self.packages_path / self.author / "skills" / skill_name).exists()
        assert f"skill/{self.author}/{skill_name}/0.1.0" in json.loads(
            (self.packages_path / PACKAGES_FILE).read_text()
        ).get("dev")
        cli_args = [
            "-m",
            "aea.cli",
            "test",
            "by-path",
            str(self.packages_path / self.author / "skills" / skill_name),
        ]

        # we use a subprocess rather than click.CliRunner because we want to isolate
        # the Python environment of the current pytest process with the subcall to pytest.main
        # from the AEA command `aea test by-path ...`
        result = self.start_subprocess(*cli_args)
        result.wait(timeout=60.0)
        assert (
            result.returncode == 0
        ), f"stdout: {self.stdout[result.pid]}, stderr: {self.stderr[result.pid]}"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Log streams"""


import re
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Generator, List, Optional, TextIO, Tuple, cast

from autonomy.analyse.logs.base import (
    ENTER_BEHAVIOUR_REGEX,
    ENTER_ROUND_REGEX,
    LOG_ROW_REGEX,
    LogRow,
    TIMESTAMP_REGEX,
    TIME_FORMAT,
)
from autonomy.analyse.logs.db import AgentLogsDB


class LogCollection(ABC):
    """Collection of logs."""

    def __init__(self) -> None:
        """Initialize object."""

        self.agents = self.get_avilable_agents()
        self.n_agents = len(self.agents)

    @abstractmethod
    def get_avilable_agents(self) -> List[str]:
        """Returns a list of agent names."""

    @abstractmethod
    def create_agent_db(
        self,
        agent: str,
        db: AgentLogsDB,
        reset: bool = False,
    ) -> "LogCollection":
        """Create logs database."""

    @staticmethod
    def get_next_log_block(
        fp: TextIO, prev_line: str
    ) -> Tuple[Optional[str], Optional[str]]:
        """Get next log block."""

        line = prev_line
        while True:
            _line = fp.readline()
            if _line == "":
                return None, None
            if TIMESTAMP_REGEX.match(_line) is not None:
                return line, _line
            line += line

    @classmethod
    def parse(cls, file: Path) -> Generator[LogRow, None, None]:
        """Parse logs and yield rows."""
        with file.open(mode="r") as fp:
            prev_line: Optional[str] = fp.readline()
            current_period = 0
            current_round = "agent_startup"
            current_behaviour = "agent_startup"
            while True:
                line, prev_line = cls.get_next_log_block(
                    fp=fp, prev_line=cast(str, prev_line)
                )
                if line is None and prev_line is None:
                    break

                match = LOG_ROW_REGEX.match(string=cast(str, line))
                _timestamp, log_level, _, log_block = cast(re.Match, match).groups()
                timestamp = datetime.strptime(_timestamp, TIME_FORMAT)

                match = ENTER_BEHAVIOUR_REGEX.match(string=log_block)
                if match is not None:
                    (current_behaviour,) = match.groups()

                match = ENTER_ROUND_REGEX.match(string=log_block)
                if match is not None:
                    (current_round, current_period) = cast(
                        Tuple[str, int], match.groups()
                    )

                yield timestamp, log_level, log_block, current_period, current_round, current_behaviour


class FromDirectory(LogCollection):
    """Log stream from directory."""

    def __init__(self, directory: Path) -> None:
        """Initialize object."""
        self.directory = directory

        super().__init__()

    def get_avilable_agents(self) -> List[str]:
        """Returns a list of agent names."""
        return list(
            map(lambda x: x.name.replace(".txt", ""), self.directory.glob("aea_*.txt"))
        )

    def create_agent_db(
        self,
        agent: str,
        db: AgentLogsDB,
        reset: bool = False,
    ) -> "FromDirectory":
        """Create logs table for agent."""

        log_file = self.directory / f"{agent}.txt"
        db.create(reset=reset)
        db.insert_many(
            logs=self.parse(
                file=log_file,
            )
        )
        return self



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test dialogues module for abci protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from aea.test_tools.test_protocol import BaseProtocolDialoguesTestCase

from packages.valory.protocols.abci.dialogues import AbciDialogue, AbciDialogues
from packages.valory.protocols.abci.message import AbciMessage


class TestDialoguesAbci(BaseProtocolDialoguesTestCase):
    """Test for the 'abci' protocol dialogues."""

    MESSAGE_CLASS = AbciMessage

    DIALOGUE_CLASS = AbciDialogue

    DIALOGUES_CLASS = AbciDialogues

    ROLE_FOR_THE_FIRST_MESSAGE = AbciDialogue.Role.CLIENT  # CHECK

    def make_message_content(self) -> dict:
        """Make a dict with message contruction content for dialogues.create."""
        return dict(
            performative=AbciMessage.Performative.REQUEST_ECHO,
            message="some str",
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Scaffold connection and channel."""
from typing import Any, Optional

from aea.configurations.base import PublicId
from aea.connections.base import BaseSyncConnection, Connection
from aea.mail.base import Envelope


"""
Choose one of the possible implementations:

Sync (inherited from BaseSyncConnection) or Async (inherited from Connection) connection and remove unused one.
"""

CONNECTION_ID = PublicId.from_str("dummy_author/dummy_connection:0.1.0")


class MyScaffoldAsyncConnection(Connection):
    """Proxy to the functionality of the SDK or API."""

    connection_id = CONNECTION_ID

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize the connection.

        The configuration must be specified if and only if the following
        parameters are None: connection_id, excluded_protocols or restricted_to_protocols.

        Possible keyword arguments:
        - configuration: the connection configuration.
        - data_dir: directory where to put local files.
        - identity: the identity object held by the agent.
        - crypto_store: the crypto store for encrypted communication.
        - restricted_to_protocols: the set of protocols ids of the only supported protocols for this connection.
        - excluded_protocols: the set of protocols ids that we want to exclude for this connection.

        :param kwargs: keyword arguments passed to component base
        """
        super().__init__(**kwargs)  # pragma: no cover

    async def connect(self) -> None:
        """
        Set up the connection.

        In the implementation, remember to update 'connection_status' accordingly.
        """
        raise NotImplementedError  # pragma: no cover

    async def disconnect(self) -> None:
        """
        Tear down the connection.

        In the implementation, remember to update 'connection_status' accordingly.
        """
        raise NotImplementedError  # pragma: no cover

    async def send(self, envelope: Envelope) -> None:
        """
        Send an envelope.

        :param envelope: the envelope to send.
        """
        raise NotImplementedError  # pragma: no cover

    async def receive(self, *args: Any, **kwargs: Any) -> Optional[Envelope]:
        """
        Receive an envelope. Blocking.

        :param args: arguments to receive
        :param kwargs: keyword arguments to receive
        :return: the envelope received, if present.  # noqa: DAR202
        """
        raise NotImplementedError  # pragma: no cover


class MyScaffoldSyncConnection(BaseSyncConnection):
    """Proxy to the functionality of the SDK or API."""

    MAX_WORKER_THREADS = 5

    connection_id = CONNECTION_ID

    def __init__(self, *args: Any, **kwargs: Any) -> None:  # pragma: no cover
        """
        Initialize the connection.

        The configuration must be specified if and only if the following
        parameters are None: connection_id, excluded_protocols or restricted_to_protocols.

        Possible arguments:
        - configuration: the connection configuration.
        - data_dir: directory where to put local files.
        - identity: the identity object held by the agent.
        - crypto_store: the crypto store for encrypted communication.
        - restricted_to_protocols: the set of protocols ids of the only supported protocols for this connection.
        - excluded_protocols: the set of protocols ids that we want to exclude for this connection.

        :param args: arguments passed to component base
        :param kwargs: keyword arguments passed to component base
        """
        super().__init__(*args, **kwargs)
        raise NotImplementedError

    def main(self) -> None:
        """
        Run synchronous code in background.

        SyncConnection `main()` usage:
        The idea of the `main` method in the sync connection
        is to provide for a way to actively generate messages by the connection via the `put_envelope` method.

        A simple example is the generation of a message every second:
        ```
        while self.is_connected:
            envelope = make_envelope_for_current_time()
            self.put_enevelope(envelope)
            time.sleep(1)
        ```
        In this case, the connection will generate a message every second
        regardless of envelopes sent to the connection by the agent.
        For instance, this way one can implement periodically polling some internet resources
        and generate envelopes for the agent if some updates are available.
        Another example is the case where there is some framework that runs blocking
        code and provides a callback on some internal event.
        This blocking code can be executed in the main function and new envelops
        can be created in the event callback.
        """
        raise NotImplementedError  # pragma: no cover

    def on_send(self, envelope: Envelope) -> None:
        """
        Send an envelope.

        :param envelope: the envelope to send.
        """
        raise NotImplementedError  # pragma: no cover

    def on_connect(self) -> None:
        """
        Tear down the connection.

        Connection status set automatically.
        """
        raise NotImplementedError  # pragma: no cover

    def on_disconnect(self) -> None:
        """
        Tear down the connection.

        Connection status set automatically.
        """
        raise NotImplementedError  # pragma: no cover



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Helper functions for checking compliance to ABCI spec"""

# pylint: skip-file

import builtins
import functools
import inspect
from enum import Enum
from pathlib import Path
from types import ModuleType
from typing import Any, Dict, List, Tuple, Type, Union

import yaml
from aea.protocols.generator.common import (
    SPECIFICATION_COMPOSITIONAL_TYPES,
    _camel_case_to_snake_case,
    _get_sub_types_of_compositional_types,
    _to_camel_case,
)
from google.protobuf.descriptor import FieldDescriptor

from packages.valory.connections.abci.connection import PUBLIC_ID
from packages.valory.connections.abci.connection import (
    _TendermintProtocolDecoder as Decoder,
)
from packages.valory.connections.abci.connection import (
    _TendermintProtocolEncoder as Encoder,
)
from packages.valory.connections.abci.dialogues import AbciDialogues
from packages.valory.connections.abci.tendermint.abci.types_pb2 import (  # type: ignore
    DESCRIPTOR,
    Request,
    Response,
)
from packages.valory.protocols import abci as valory_abci_protocol
from packages.valory.protocols.abci import AbciMessage, custom_types


Node = Dict[str, Any]

camel_to_snake = _camel_case_to_snake_case
snake_to_camel = _to_camel_case


type_mapping = {
    v: k[5:].lower() for k, v in vars(FieldDescriptor).items() if k.startswith("TYPE_")
}

type_to_python = dict.fromkeys(type_mapping.values())
type_to_python.update(
    dict(
        double=float,
        float=float,
        int64=int,
        uint64=int,
        int32=int,
        bool=bool,
        string=str,
        bytes=bytes,
        uint32=int,
    )
)

TENDERMINT_PRIMITIVES = tuple(type_mapping.values())
PYTHON_PRIMITIVES = (int, float, bool, str, bytes)

# simple value strategy
AEA_PRIMITIVES = {str: "sss", bytes: b"bbb", int: 123, float: 3.14, bool: True}
TENDER_PRIMITIVES = dict(
    int32=-32,
    uint32=32,
    int64=-64,
    uint64=64,
    bool=True,
    string="sss",
    bytes=b"bbb",
    nanos=999,
)
REPEATED_FIELD_SIZE = 3
USE_NON_ZERO_ENUM: bool = True


def is_enum(d_type: Any) -> bool:
    """Check if a type is an Enum (not instance!)."""
    return isinstance(d_type, type) and issubclass(d_type, Enum)


def my_repr(self: Any) -> str:
    """Custom __repr__ for Tendermint protobuf objects, which lack it."""
    return f"<{self.__module__}.{type(self).__name__} object at {hex(id(self))}>"


def is_typing_list(d_type: Any) -> bool:
    """Check if field is repeated."""
    return d_type.__class__.__module__ == "typing" and d_type.__origin__ is list


def replace_keys(node: Node, trans: Node) -> None:
    """Replace keys in-place"""
    for k, v in trans.items():
        if isinstance(v, dict):
            replace_keys(node[k], v)
        else:
            node[v] = node.pop(k)


def set_repr(cls: Type) -> Type:
    """Set custom __repr__"""
    cls.__repr__ = my_repr  # type: ignore
    return cls


def get_aea_classes(module: ModuleType) -> Dict[str, Type]:
    """Get AEA custom classes."""

    def is_locally_defined_class(item: Any) -> bool:
        return isinstance(item, type) and item.__module__ == module.__name__

    return {k: v for k, v in vars(module).items() if is_locally_defined_class(v)}


AEA_CUSTOM = get_aea_classes(custom_types)


@functools.lru_cache()
def get_protocol_readme_spec() -> Tuple[Any, Any, Any]:
    """Test specification used to generate protocol matches ABCI spec"""

    protocol_readme = Path(valory_abci_protocol.__file__).parent / "README.md"
    raw_chunks = open(protocol_readme).read().split("```")
    assert len(raw_chunks) == 3, "Expecting a single YAML code block"

    yaml_chunks = raw_chunks[1].strip("yaml").split("\n---")
    yaml_content = []
    for raw_doc in filter(None, yaml_chunks):
        try:
            yaml_content.append(yaml.safe_load(raw_doc))
        except yaml.YAMLError as e:
            raise e

    protocol, custom, dialogues = yaml_content
    assert protocol["name"] == "abci"  # sanity check
    return protocol, custom, dialogues


# 1. Create type tree for the AEA-native ABCI protocol
def _create_aea_custom_type_tree(custom_type: Type) -> Tuple[Type, Node]:
    """Create custom type tree for AEA-native ABCI spec"""

    kwarg_types = {}
    parameters = inspect.signature(custom_type).parameters
    for name, parameter in parameters.items():
        d_type = parameter.annotation
        if d_type in PYTHON_PRIMITIVES:
            kwarg_types[name] = d_type
        elif is_enum(d_type):
            kwarg_types[name] = d_type
        elif is_typing_list(d_type):  # "repeated"
            assert len(d_type.__args__) == 1  # check assumption
            container, content = d_type.__origin__, d_type.__args__[0]
            if content in AEA_CUSTOM.values():
                content = _create_aea_custom_type_tree(content)
            kwarg_types[name] = container, content
        else:
            nested_type = AEA_CUSTOM.get(d_type, d_type)
            kwarg_types[name] = _create_aea_custom_type_tree(nested_type)

    return custom_type, kwarg_types


def _create_aea_type_tree(field: str) -> Any:
    """Create type tree for AEA-native ABCI spec"""

    if any(map(field.startswith, SPECIFICATION_COMPOSITIONAL_TYPES)):
        subfields = _get_sub_types_of_compositional_types(field)
        if field.startswith("pt:optional"):
            return _create_aea_type_tree(*subfields)
        elif field.startswith("pt:list"):  # repeated
            return list, _create_aea_type_tree(*subfields)
        raise NotImplementedError(f"field: {field}")

    if field.startswith("ct:"):
        custom_type = AEA_CUSTOM[field[3:]]
        return _create_aea_custom_type_tree(custom_type)

    primitive = getattr(builtins, field[3:])
    return primitive


def create_aea_abci_type_tree(
    speech_acts: Dict[str, Dict[str, str]]
) -> Dict[str, Node]:
    """Create AEA-native ABCI type tree from the defined speech acts"""

    def _get_message_types(fields: Dict[str, str]) -> Node:
        """Get message types for AEA-native ABCI spec"""
        return {k: _create_aea_type_tree(v) for k, v in fields.items()}

    return {k: _get_message_types(v) for k, v in speech_acts.items()}


# 2. Initialize AEA-native ABCI primitives
def _init_subtree(node: Node) -> Node:
    """Initialize subtree of type_tree non-custom objects"""

    def init_repeated(repeated_type: Any) -> Tuple[Any, ...]:
        """Repeated fields must be tuples for Tendermint protobuf"""

        if repeated_type in PYTHON_PRIMITIVES:
            data = AEA_PRIMITIVES[repeated_type]
            return tuple(data for _ in range(REPEATED_FIELD_SIZE))
        elif isinstance(repeated_type, tuple):
            cls, kws = repeated_type
            return tuple(_init_subtree(kws) for _ in range(REPEATED_FIELD_SIZE))
        raise NotImplementedError(f"Repeated in {name}: {repeated_type}")

    kwargs: Node = {}
    for name, d_type in node.items():
        if isinstance(d_type, tuple):
            container, content = d_type
            if container is list:
                kwargs[name] = init_repeated(content)
            else:
                kwargs[name] = _init_subtree(content)
        elif d_type in PYTHON_PRIMITIVES:
            kwargs[name] = AEA_PRIMITIVES[d_type]
        elif is_enum(d_type):
            kwargs[name] = list(d_type)[USE_NON_ZERO_ENUM]
        else:
            raise NotImplementedError(f"{name}: {d_type}")

    return kwargs


def init_type_tree_primitives(type_tree: Node) -> Node:
    """
    Initialize the primitive types and size of repeated fields.

    These are the only initialization parameters that can vary;
    after this the initialization of custom types is what remains

    This structure allows:
    - Comparison of structure and these values with Tendermint translation
    - Visual inspection of fields to be sets, also on custom objects
    - Randomized testing strategies using e.g. hypothesis

    :param type_tree: mapping from message / field name to type.
    :return: mapping from message / field name to initialized primitive.
    """
    return {k: _init_subtree(node) for k, node in type_tree.items()}


# 3. Initialize AEA-native ABCI protocol messages
def _complete_init(type_node: Node, init_node: Node) -> Node:
    """Initialize custom classes and containers"""

    def init_repeated(repeated_type: Any) -> Tuple[Any, ...]:
        if not isinstance(content, tuple):
            return init_node[key]  # already initialized primitives
        custom_type, kwargs = repeated_type
        return tuple(custom_type(**_complete_init(kwargs, c)) for c in init_node[key])

    node: Node = {}
    for key, d_type in type_node.items():
        if d_type in PYTHON_PRIMITIVES:
            node[key] = init_node[key]
        elif isinstance(d_type, tuple):
            container, content = d_type
            if container is list:
                node[key] = init_repeated(content)
            else:
                node[key] = container(**_complete_init(content, init_node[key]))
        elif is_enum(d_type):
            node[key] = init_node[key]
        else:
            raise NotImplementedError(f"{key}: {d_type}")

    return node


def init_aea_abci_messages(type_tree: Node, init_tree: Node) -> Node:
    """
    Create ABCI messages for AEA-native ABCI spec

    We iterate the type_tree and init_tree to finalize the
    initialization of custom objects contained in it, and
    create an instance of all ABCI messages.

    :param type_tree: mapping from message / field name to type.
    :param init_tree: mapping from message / field name to initialized primitive.
    :return: mapping from message name to ABCI Message instance
    """

    messages = dict.fromkeys(type_tree)
    for key in type_tree:
        performative = getattr(AbciMessage.Performative, key.upper())
        kwargs = _complete_init(type_tree[key], init_tree[key])
        messages[key] = AbciMessage(performative, **kwargs)

    return messages


class EncodingError(Exception):
    """EncodingError AEA- to Tendermint-native ABCI message"""


class DecodingError(Exception):
    """DecodingError Tendermint- to AEA-native ABCI message"""


# 4. Translate AEA-native to Tendermint-native
def encode(message: AbciMessage) -> Response:
    """Encode AEA-native ABCI protocol messages to Tendermint-native"""

    try:
        return Encoder.process(message)
    except Exception as e:
        raise EncodingError(f"ABCI message {message}: {e}")


def decode(request: Request) -> AbciMessage:
    """Decode Tendermint-native ABCI protocol messages to AEA-native"""

    dialogues = AbciDialogues(connection_id=PUBLIC_ID)
    try:
        message, dialogue = Decoder().process(request, dialogues, "dummy")  # type: ignore
        return message
    except Exception as e:
        raise DecodingError(f"Request {request}: {e}")


# 5. Build content tree from Tendermint-native ABCI messages
def _get_message_content(message: Any) -> Node:
    """Verify Tendermint-native ABCI message"""

    # NOTE: PublicKey objects are an Enum in AEA-native protocol
    #       but are retrieved as mapping from Tendermint side

    assert message.IsInitialized()
    assert not message.UnknownFields()
    assert not message.FindInitializationErrors()

    # NOTE: ListFields does not retrieve what is empty!
    fields = dict(message.DESCRIPTOR.fields_by_name)
    enum_types = dict(message.DESCRIPTOR.enum_types_by_name)
    oneofs = dict(message.DESCRIPTOR.oneofs_by_name)

    kwargs: Node = {}
    for name, descr in fields.items():
        attr = getattr(message, name)

        if descr.enum_type:
            enum_type = enum_types.pop(snake_to_camel(name))
            kwargs[name] = (enum_type.values[attr].name, attr)
        elif isinstance(attr, PYTHON_PRIMITIVES):
            kwargs[name] = attr
        elif descr.label == descr.LABEL_REPEATED:
            assert isinstance(descr.default_value, list)  # attr
            assert len(set(map(type, attr))) <= 1, "Expecting single type"
            if not attr or isinstance(attr[0], PYTHON_PRIMITIVES):
                kwargs[name] = list(attr)
            else:
                kwargs[name] = [_get_message_content(c) for c in attr]
        elif descr.message_type:
            kwargs[name] = _get_message_content(attr)
        else:
            raise NotImplementedError(f"name: {name} {attr}")

    # in the case of oneofs, we assert they were retrieved from fields (PublicKey)
    expected = {f.name for oneof in oneofs.values() for f in oneof.fields}
    assert all(k in kwargs for k in expected), f"oneofs in {message}"
    assert not enum_types, f"assumed max 1 enum per message: {message}"

    return kwargs


def get_tendermint_content(envelope: Union[Request, Response]) -> Node:
    """
    Get Tendermint-native ABCI message content.

    For all Request / Response instances obtained after encoding,
    we retrieve the information present in the message they contain.

    :param envelope: a Tendermint Request / Response object.
    :return: mapping structure from message / field name to leaf values
    """

    assert isinstance(envelope, (Request, Response))
    assert envelope.IsInitialized()
    assert not envelope.UnknownFields()
    assert not envelope.FindInitializationErrors()
    assert len(envelope.ListFields()) == 1
    descr, message = envelope.ListFields()[0]
    return _get_message_content(message)


# compare AEA- and Tendermint-native ABCI protocol input and output
def compare_trees(init_node: Node, tender_node: Node) -> None:
    """Compare Initialization and Tendermint tree nodes"""

    # NOTE: PublicKey objects are an Enum in AEA-native protocol
    #       but are retrieved as mapping from Tendermint side

    if init_node == tender_node:
        return

    for k, init_child in init_node.items():

        # translate key to tendermint key
        tk = {"type_": "type", "format_": "format", "hash_": "hash"}.get(k, k)
        tender_child = tender_node[tk]

        if k == "pub_key":
            init_child = {init_child["key_type"].name: init_child["data"]}

        if init_child == tender_child:
            continue
        elif isinstance(init_child, dict) and isinstance(tender_child, dict):
            compare_trees(init_child, tender_child)

        elif isinstance(tender_child, list):

            # we use a nested mapping to represent a custom class,
            # tendermint doesn't use this for storing repeated fields
            if isinstance(init_child, dict):
                repeated = list(init_child.values())
                assert len(repeated) == 1
                init_child = repeated[0]

            assert len(init_child) == len(tender_child)
            for a, b in zip(init_child, tender_child):
                compare_trees(a, b)

        else:
            assert len(init_child) == 1, "expecting enum"
            enum = set(init_child.values()).pop()
            assert (enum.name, enum.value) == tender_node[tk]


def _process_message_descriptor(m_descriptor: Any) -> Node:
    """Process fields of the message descriptor"""

    node: Node = {}
    for name, field in m_descriptor.fields_by_name.items():

        if field.message_type:
            cls = set_repr(field.message_type._concrete_class)
            node[name] = cls, _process_message_descriptor(cls.DESCRIPTOR)
        elif field.enum_type:
            names = field.enum_type.values_by_name
            numbers = field.enum_type.values_by_number
            node[name] = Enum(name, zip(names, numbers))
        else:
            node[name] = type_mapping[field.type]

        if field.label == field.LABEL_REPEATED:
            node[name] = list, node[name]

    return node


def get_tender_type_tree() -> Node:
    """Tendermint type tree"""

    descriptor = DESCRIPTOR
    tender_type_tree = {}
    for msg, msg_descr in descriptor.message_types_by_name.items():
        cls = set_repr(msg_descr._concrete_class)
        tender_type_tree[msg] = cls, _process_message_descriptor(msg_descr)
    return tender_type_tree


def _init_tender_tree(tender_node: Node) -> Node:
    """Initialize Tendermint classes"""

    def init_repeated(repeated_type: Any) -> List[Any]:
        if isinstance(repeated_type, str):
            return [
                TENDER_PRIMITIVES[repeated_type] for _ in range(REPEATED_FIELD_SIZE)
            ]
        cls, kwargs = repeated_type
        return [cls(**_init_tender_tree(kwargs)) for _ in range(REPEATED_FIELD_SIZE)]

    node = {}
    for field, d_type in tender_node.items():
        if isinstance(d_type, str):
            if field == "nanos":  # special
                node[field] = TENDER_PRIMITIVES[field]
            else:
                node[field] = TENDER_PRIMITIVES[d_type]
        elif is_enum(d_type):
            node[field] = list(d_type)[USE_NON_ZERO_ENUM].value
        elif isinstance(d_type, tuple):
            container, content = d_type
            if container is list:
                node[field] = init_repeated(content)
            else:
                node[field] = container(**_init_tender_tree(content))
        else:
            raise NotImplementedError(f"field {field}: {d_type}")

    return node


def _build_tendermint_messages(cls: Type, tree: Node) -> List[Union[Request, Response]]:
    """Build tendermint messages"""

    return [cls(**{k: v}) for k, v in _init_tender_tree(tree).items()]


def init_tendermint_messages(tender_type_tree: Node) -> List[Union[Request, Response]]:
    """Initialize tendermint ABCI messages"""

    request_cls, request_tree = tender_type_tree["Request"]

    # construct Tendermint-native messages
    messages = _build_tendermint_messages(request_cls, request_tree)

    return messages



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Core for cli."""

import click
from aea.cli.core import cli

import autonomy

# trigger patching of 'aea scaffold' command
import autonomy.cli.scaffold_fsm  # noqa
from autonomy.cli.analyse import analyse_group
from autonomy.cli.build_images import build_image
from autonomy.cli.deploy import deploy_group
from autonomy.cli.develop import develop_group
from autonomy.cli.fetch import fetch
from autonomy.cli.hash import hash_group
from autonomy.cli.mint import mint
from autonomy.cli.packages import package_manager
from autonomy.cli.publish import publish
from autonomy.cli.push_all import push_all
from autonomy.cli.replay import replay_group
from autonomy.cli.service import service


cli.add_command(analyse_group)
cli.add_command(deploy_group)
cli.add_command(develop_group)
cli.add_command(replay_group)
cli.add_command(hash_group)
cli.add_command(push_all)
cli.add_command(publish)
cli.add_command(fetch)
cli.add_command(build_image)
cli.add_command(package_manager)
cli.add_command(mint)
cli.add_command(service)


click.version_option(autonomy.__version__, prog_name="autonomy")(cli)
cli.help = "Command-line tool for managing agent services of the Open Autonomy framework.\n\nExtends the command-line tool for setting up an Autonomous Economic Agent (AEA)."
cli.name = "autonomy"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tools for analysing logs."""


import re
from datetime import datetime
from typing import Tuple


LOGS_DB = "logs.db"
TIME_FORMAT = "%Y-%m-%d %H:%M:%S,%f"
TIMESTAMP_REGEX = re.compile(r"^(\[(\d+-\d+-\d+ \d+:\d+:\d+,\d+)\])")
LOG_ROW_REGEX = re.compile(
    r"\[(\d+-\d+-\d+ \d+:\d+:\d+,\d+)\] \[([A-Z]+)\]( \[agent\])? (.*)"
)
ENTER_BEHAVIOUR_REGEX = re.compile(r"Entered in the \'([a-z_]+)\' behaviour")
ENTER_ROUND_REGEX = re.compile(r"Entered in the \'([a-z_]+)\' round for period (\d+)")
EXIT_ROUND_REGEX = re.compile(r"'([a-z_]+)' round is done with event: (Event\.[A-Z_]+)")

LogRow = Tuple[datetime, str, str, int, str, str]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""IPFS hash helpers."""

import traceback
from pathlib import Path
from typing import Callable, Dict, Optional, cast

import click
from aea.cli.ipfs_hash import (
    extend_public_ids,
    hash_package,
    package_id_and_path,
    sort_configuration_file,
    to_package_id,
)
from aea.configurations.base import PackageConfiguration, PackageType
from aea.configurations.constants import PACKAGE_TYPE_TO_CONFIG_FILE, SCAFFOLD_PACKAGES
from aea.configurations.data_types import PackageId, PublicId
from aea.configurations.loader import load_configuration_object
from aea.helpers.dependency_tree import DependencyTree, dump_yaml, load_yaml
from aea.helpers.fingerprint import update_fingerprint

from autonomy.configurations.base import PACKAGE_TYPE_TO_CONFIG_CLASS


def load_configuration(
    package_type: PackageType, package_path: Path
) -> PackageConfiguration:
    """
    Load a configuration, knowing the type and the path to the package root.

    :param package_type: the package type.
    :param package_path: the path to the package root.
    :return: the configuration object.
    """
    configuration_obj = load_configuration_object(
        package_type,
        package_path,
        package_type_config_class=PACKAGE_TYPE_TO_CONFIG_CLASS,
    )
    configuration_obj._directory = package_path  # pylint: disable=protected-access
    return cast(PackageConfiguration, configuration_obj)


# TODO: extract into utils
# Add input validations
def update_hashes(  # pylint: disable=too-many-locals
    packages_dir: Path,
    no_wrap: bool = False,
    vendor: Optional[str] = None,
    config_loader: Callable[
        [PackageType, Path], PackageConfiguration
    ] = load_configuration,
) -> int:
    """Process all AEA packages, update fingerprint, and update packages.json file."""
    return_code = 0
    package_hashes: Dict[str, str] = {}

    try:
        public_id_to_hash_mappings: Dict = {}
        dependency_tree = DependencyTree.generate(packages_dir)
        packages = [
            [package_id_and_path(package_id, packages_dir) for package_id in tree_level]
            for tree_level in dependency_tree
        ]
        packages[0] = packages[0] + list(map(to_package_id, SCAFFOLD_PACKAGES))
        for tree_level in packages:
            for package_id, package_path in tree_level:
                click.echo(
                    "Processing package {} of type {}".format(
                        package_path.name, package_id.package_type
                    )
                )

                config_file = package_path / cast(
                    str, PACKAGE_TYPE_TO_CONFIG_FILE.get(package_id.package_type.value)
                )
                item_config, extra_config = load_yaml(config_file)
                if package_id.package_type == PackageType.SERVICE:
                    agent_id = PackageId(
                        PackageType.AGENT, PublicId.from_str(item_config["agent"])
                    )
                    item_config["agent"] = str(
                        PublicId(
                            author=agent_id.author,
                            name=agent_id.name,
                            version=agent_id.version,
                            package_hash=public_id_to_hash_mappings[agent_id],
                        )
                    )
                else:
                    extend_public_ids(item_config, public_id_to_hash_mappings)

                dump_yaml(config_file, item_config, extra_config)

                configuration_obj = config_loader(
                    package_id.package_type.value, package_path
                )
                sort_configuration_file(configuration_obj)
                update_fingerprint(configuration_obj)
                key, package_hash = hash_package(
                    configuration_obj, package_id.package_type, no_wrap=no_wrap
                )
                public_id_to_hash_mappings[package_id] = package_hash

                if vendor is not None and package_id.author != vendor:
                    continue  # pragma: nocover
                package_hashes[key] = package_hash
        click.echo("Done!")

    except Exception:  # pylint: disable=broad-except  # pragma: nocover
        traceback.print_exc()
        return_code = 1

    return return_code



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test agent runner."""


import os
import shutil
from pathlib import Path
from typing import Any, Tuple
from unittest import mock

from autonomy.cli import cli
from autonomy.deploy.base import TENDERMINT_COM_URL_PARAM, TENDERMINT_URL_PARAM
from autonomy.replay.agent import AgentRunner

from tests.conftest import ROOT_DIR, skip_docker_tests
from tests.test_autonomy.test_cli.base import BaseCliTest


OS_ENV_PATCH = mock.patch.dict(
    os.environ, values={**os.environ, "ALL_PARTICIPANTS": "[]"}, clear=True
)

DOCKER_COMPOSE_DATA = {
    "version": "2.4",
    "services": {
        "abci0": {
            "mem_limit": "1024m",
            "mem_reservation": "256M",
            "cpus": 1,
            "container_name": "abci0",
            "image": "valory/open-autonomy-open-aea:oracle_deployable-0.1.0",
            "environment": [
                "ID=DUMMY",
                "LOG_FILE=DUMMY",
                "AEA_KEY=DUMMY",
                "VALORY_APPLICATION=DUMMY",
                "ABCI_HOST=DUMMY",
                f"SKILL_ORACLE_ABCI_MODELS_PARAMS_ARGS_{TENDERMINT_URL_PARAM.upper()}=DUMMY",
                f"SKILL_ORACLE_ABCI_MODELS_PARAMS_ARGS_{TENDERMINT_COM_URL_PARAM.upper()}=DUMMY",
                "SKILL_ORACLE_ABCI_MODELS_PRICE_API_ARGS_URL=DUMMY",
                "SKILL_ORACLE_ABCI_MODELS_PRICE_API_ARGS_API_ID=DUMMY",
                "SKILL_ORACLE_ABCI_MODELS_PRICE_API_ARGS_PARAMETERS=DUMMY",
                "SKILL_ORACLE_ABCI_MODELS_PRICE_API_ARGS_RESPONSE_KEY=DUMMY",
                "SKILL_ORACLE_ABCI_MODELS_PRICE_API_ARGS_HEADERS=DUMMY",
                "SKILL_ORACLE_ABCI_MODELS_RANDOMNESS_API_ARGS_URL=DUMMY",
                "SKILL_ORACLE_ABCI_MODELS_RANDOMNESS_API_ARGS_API_ID=DUMMY",
                "SKILL_ORACLE_ABCI_MODELS_PARAMS_ARGS_RESET_PAUSE_DURATION=DUMMY",
                "SKILL_ORACLE_ABCI_MODELS_PARAMS_ARGS_BROADCAST_TO_SERVER=DUMMY",
                "SKILL_ORACLE_ABCI_MODELS_SERVER_API_ARGS_URL=DUMMY",
                "SKILL_ORACLE_ABCI_MODELS_BENCHMARK_TOOL_ARGS_LOG_DIR=DUMMY",
                "LEDGER_ADDRESS=DUMMY",
                "LEDGER_CHAIN_ID=DUMMY",
            ],
        }
    },
}


def ctrl_c(*args: Any) -> None:
    """Send control C."""

    raise KeyboardInterrupt()


@skip_docker_tests
class TestAgentRunner(BaseCliTest):
    """Test agent runner tool."""

    cli_options: Tuple[str, ...] = ("replay", "agent")
    packages_dir: Path = ROOT_DIR / "packages"
    output_dir: Path = ROOT_DIR
    keys_path: Path = ROOT_DIR / "deployments" / "keys" / "hardhat_keys.json"

    def setup(self) -> None:
        """Setup test method."""
        super().setup()

        shutil.copytree(
            self.packages_dir / "valory" / "services" / "hello_world",
            self.t / "hello_world",
        )
        os.chdir(self.t / "hello_world")

    def test_run(self) -> None:
        """Test run."""

        with mock.patch("os.chown"), OS_ENV_PATCH:
            result = self.cli_runner.invoke(
                cli,
                (
                    "deploy",
                    "build",
                    str(self.keys_path),
                    "--local",
                    "--o",
                    str(self.t / "abci_build"),
                ),
            )

        assert result.exit_code == 0, result.output

        build_dir = self.t / "abci_build"
        with mock.patch.object(AgentRunner, "start", new=ctrl_c), mock.patch.object(
            AgentRunner, "stop"
        ) as stop_mock, mock.patch(
            "autonomy.cli.replay.load_docker_config", new=lambda x: DOCKER_COMPOSE_DATA
        ):
            result = self.run_cli(
                ("--registry", str(self.packages_dir), "0", "--build", str(build_dir))
            )
            assert result.exit_code == 0, result.output
            stop_mock.assert_any_call()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Conftest module for io tests."""

import os
import shutil
from contextlib import suppress
from pathlib import Path
from typing import Dict, Generator

import pytest
from hypothesis import settings

from packages.valory.skills.abstract_round_abci.io_.store import StoredJSONType
from packages.valory.skills.abstract_round_abci.models import MIN_RESET_PAUSE_DURATION


# pylint: skip-file


CI = "CI"
PACKAGE_DIR = Path(__file__).parent.parent
settings.register_profile(CI, deadline=5000)
profile_name = ("default", "CI")[bool(os.getenv("CI"))]


@pytest.fixture
def dummy_obj() -> StoredJSONType:
    """A dummy custom object to test the storing with."""
    return {"test_col": ["test_val_1", "test_val_2"]}


@pytest.fixture
def dummy_multiple_obj(dummy_obj: StoredJSONType) -> Dict[str, StoredJSONType]:
    """Many dummy custom objects to test the storing with."""
    return {f"test_obj_{i}": dummy_obj for i in range(10)}


@pytest.fixture(scope="session", autouse=True)
def hypothesis_cleanup() -> Generator:
    """Fixture to remove hypothesis directory after tests."""
    yield
    hypothesis_dir = PACKAGE_DIR / ".hypothesis"
    if hypothesis_dir.exists():
        with suppress(OSError, PermissionError):  # pragma: nocover
            shutil.rmtree(hypothesis_dir)


# We do not care about these keys but need to set them in the behaviours' tests,
# because `packages.valory.skills.abstract_round_abci.models._ensure` is used.
irrelevant_genesis_config = {
    "consensus_params": {
        "block": {"max_bytes": "str", "max_gas": "str", "time_iota_ms": "str"},
        "evidence": {
            "max_age_num_blocks": "str",
            "max_age_duration": "str",
            "max_bytes": "str",
        },
        "validator": {"pub_key_types": ["str"]},
        "version": {},
    },
    "genesis_time": "str",
    "chain_id": "str",
    "voting_power": "str",
}
irrelevant_config = {
    "tendermint_url": "str",
    "max_healthcheck": 0,
    "round_timeout_seconds": 0.0,
    "sleep_time": 0,
    "retry_timeout": 0,
    "retry_attempts": 0,
    "keeper_timeout": 0.0,
    "reset_pause_duration": MIN_RESET_PAUSE_DURATION,
    "drand_public_key": "str",
    "tendermint_com_url": "str",
    "tendermint_max_retries": 0,
    "reset_tendermint_after": 0,
    "cleanup_history_depth": 0,
    "voting_power": 0,
    "tendermint_check_sleep_delay": 0,
    "cleanup_history_depth_current": None,
    "request_timeout": 0.0,
    "request_retry_delay": 0.0,
    "tx_timeout": 0.0,
    "max_attempts": 0,
    "service_registry_address": None,
    "on_chain_service_id": None,
    "share_tm_config_on_startup": False,
    "tendermint_p2p_url": "str",
    "setup": {},
    "genesis_config": irrelevant_genesis_config,
    "use_termination": False,
}



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Custom objects for the transaction settlement ABCI application."""
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from web3.types import Nonce, Wei

from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.skills.abstract_round_abci.models import ApiSpecs, BaseParams
from packages.valory.skills.abstract_round_abci.models import (
    BenchmarkTool as BaseBenchmarkTool,
)
from packages.valory.skills.abstract_round_abci.models import Requests as BaseRequests
from packages.valory.skills.abstract_round_abci.models import (
    SharedState as BaseSharedState,
)
from packages.valory.skills.abstract_round_abci.models import TypeCheckMixin
from packages.valory.skills.transaction_settlement_abci.rounds import (
    TransactionSubmissionAbciApp,
)


_MINIMUM_VALIDATE_TIMEOUT = 300  # 5 minutes
BenchmarkTool = BaseBenchmarkTool


class SharedState(BaseSharedState):
    """Keep the current shared state of the skill."""

    abci_app_cls = TransactionSubmissionAbciApp


@dataclass
class MutableParams(TypeCheckMixin):
    """Collection for the mutable parameters."""

    fallback_gas: int
    tx_hash: str = ""
    nonce: Optional[Nonce] = None
    gas_price: Optional[Dict[str, Wei]] = None
    late_messages: List[ContractApiMessage] = field(default_factory=list)


class TransactionParams(BaseParams):  # pylint: disable=too-many-instance-attributes
    """Transaction settlement agent-specific parameters."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the parameters object.

        We keep track of the nonce and tip across rounds and periods.
        We reuse it each time a new raw transaction is generated. If
        at the time of the new raw transaction being generated the nonce
        on the ledger does not match the nonce on the skill, then we ignore
        the skill nonce and tip (effectively we price fresh). Otherwise, we
        are in a re-submission scenario where we need to take account of the
        old tip.

        :param args: positional arguments
        :param kwargs: keyword arguments
        """
        self.mutable_params = MutableParams(
            fallback_gas=self._ensure("init_fallback_gas", kwargs, int)
        )
        self.keeper_allowed_retries: int = self._ensure(
            "keeper_allowed_retries", kwargs, int
        )
        self.validate_timeout: int = self._ensure_gte(
            "validate_timeout",
            kwargs,
            int,
            min_value=_MINIMUM_VALIDATE_TIMEOUT,
        )
        self.finalize_timeout: float = self._ensure("finalize_timeout", kwargs, float)
        self.history_check_timeout: int = self._ensure(
            "history_check_timeout", kwargs, int
        )
        super().__init__(*args, **kwargs)


RandomnessApi = ApiSpecs
Requests = BaseRequests



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tendermint.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x10tendermint.proto\x12\x1c\x61\x65\x61.valory.tendermint.v0_1_0"\xad\t\n\x11TendermintMessage\x12S\n\x05\x65rror\x18\x05 \x01(\x0b\x32\x42.aea.valory.tendermint.v0_1_0.TendermintMessage.Error_PerformativeH\x00\x12\x61\n\x0cgenesis_info\x18\x06 \x01(\x0b\x32I.aea.valory.tendermint.v0_1_0.TendermintMessage.Genesis_Info_PerformativeH\x00\x12i\n\x10get_genesis_info\x18\x07 \x01(\x0b\x32M.aea.valory.tendermint.v0_1_0.TendermintMessage.Get_Genesis_Info_PerformativeH\x00\x12o\n\x13get_recovery_params\x18\x08 \x01(\x0b\x32P.aea.valory.tendermint.v0_1_0.TendermintMessage.Get_Recovery_Params_PerformativeH\x00\x12g\n\x0frecovery_params\x18\t \x01(\x0b\x32L.aea.valory.tendermint.v0_1_0.TendermintMessage.Recovery_Params_PerformativeH\x00\x1a\x8e\x01\n\tErrorCode\x12[\n\nerror_code\x18\x01 \x01(\x0e\x32G.aea.valory.tendermint.v0_1_0.TendermintMessage.ErrorCode.ErrorCodeEnum"$\n\rErrorCodeEnum\x12\x13\n\x0fINVALID_REQUEST\x10\x00\x1a\x44\n\x1dGet_Genesis_Info_Performative\x12\r\n\x05query\x18\x01 \x01(\t\x12\x14\n\x0cquery_is_set\x18\x02 \x01(\x08\x1aG\n Get_Recovery_Params_Performative\x12\r\n\x05query\x18\x01 \x01(\t\x12\x14\n\x0cquery_is_set\x18\x02 \x01(\x08\x1a)\n\x19Genesis_Info_Performative\x12\x0c\n\x04info\x18\x01 \x01(\t\x1a.\n\x1cRecovery_Params_Performative\x12\x0e\n\x06params\x18\x01 \x01(\t\x1a\x8f\x02\n\x12\x45rror_Performative\x12M\n\nerror_code\x18\x01 \x01(\x0b\x32\x39.aea.valory.tendermint.v0_1_0.TendermintMessage.ErrorCode\x12\x11\n\terror_msg\x18\x02 \x01(\t\x12\x65\n\nerror_data\x18\x03 \x03(\x0b\x32Q.aea.valory.tendermint.v0_1_0.TendermintMessage.Error_Performative.ErrorDataEntry\x1a\x30\n\x0e\x45rrorDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x0e\n\x0cperformativeb\x06proto3'
)


_TENDERMINTMESSAGE = DESCRIPTOR.message_types_by_name["TendermintMessage"]
_TENDERMINTMESSAGE_ERRORCODE = _TENDERMINTMESSAGE.nested_types_by_name["ErrorCode"]
_TENDERMINTMESSAGE_GET_GENESIS_INFO_PERFORMATIVE = (
    _TENDERMINTMESSAGE.nested_types_by_name["Get_Genesis_Info_Performative"]
)
_TENDERMINTMESSAGE_GET_RECOVERY_PARAMS_PERFORMATIVE = (
    _TENDERMINTMESSAGE.nested_types_by_name["Get_Recovery_Params_Performative"]
)
_TENDERMINTMESSAGE_GENESIS_INFO_PERFORMATIVE = _TENDERMINTMESSAGE.nested_types_by_name[
    "Genesis_Info_Performative"
]
_TENDERMINTMESSAGE_RECOVERY_PARAMS_PERFORMATIVE = (
    _TENDERMINTMESSAGE.nested_types_by_name["Recovery_Params_Performative"]
)
_TENDERMINTMESSAGE_ERROR_PERFORMATIVE = _TENDERMINTMESSAGE.nested_types_by_name[
    "Error_Performative"
]
_TENDERMINTMESSAGE_ERROR_PERFORMATIVE_ERRORDATAENTRY = (
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE.nested_types_by_name["ErrorDataEntry"]
)
_TENDERMINTMESSAGE_ERRORCODE_ERRORCODEENUM = (
    _TENDERMINTMESSAGE_ERRORCODE.enum_types_by_name["ErrorCodeEnum"]
)
TendermintMessage = _reflection.GeneratedProtocolMessageType(
    "TendermintMessage",
    (_message.Message,),
    {
        "ErrorCode": _reflection.GeneratedProtocolMessageType(
            "ErrorCode",
            (_message.Message,),
            {
                "DESCRIPTOR": _TENDERMINTMESSAGE_ERRORCODE,
                "__module__": "tendermint_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage.ErrorCode)
            },
        ),
        "Get_Genesis_Info_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Genesis_Info_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _TENDERMINTMESSAGE_GET_GENESIS_INFO_PERFORMATIVE,
                "__module__": "tendermint_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage.Get_Genesis_Info_Performative)
            },
        ),
        "Get_Recovery_Params_Performative": _reflection.GeneratedProtocolMessageType(
            "Get_Recovery_Params_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _TENDERMINTMESSAGE_GET_RECOVERY_PARAMS_PERFORMATIVE,
                "__module__": "tendermint_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage.Get_Recovery_Params_Performative)
            },
        ),
        "Genesis_Info_Performative": _reflection.GeneratedProtocolMessageType(
            "Genesis_Info_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _TENDERMINTMESSAGE_GENESIS_INFO_PERFORMATIVE,
                "__module__": "tendermint_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage.Genesis_Info_Performative)
            },
        ),
        "Recovery_Params_Performative": _reflection.GeneratedProtocolMessageType(
            "Recovery_Params_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _TENDERMINTMESSAGE_RECOVERY_PARAMS_PERFORMATIVE,
                "__module__": "tendermint_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage.Recovery_Params_Performative)
            },
        ),
        "Error_Performative": _reflection.GeneratedProtocolMessageType(
            "Error_Performative",
            (_message.Message,),
            {
                "ErrorDataEntry": _reflection.GeneratedProtocolMessageType(
                    "ErrorDataEntry",
                    (_message.Message,),
                    {
                        "DESCRIPTOR": _TENDERMINTMESSAGE_ERROR_PERFORMATIVE_ERRORDATAENTRY,
                        "__module__": "tendermint_pb2"
                        # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage.Error_Performative.ErrorDataEntry)
                    },
                ),
                "DESCRIPTOR": _TENDERMINTMESSAGE_ERROR_PERFORMATIVE,
                "__module__": "tendermint_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage.Error_Performative)
            },
        ),
        "DESCRIPTOR": _TENDERMINTMESSAGE,
        "__module__": "tendermint_pb2"
        # @@protoc_insertion_point(class_scope:aea.valory.tendermint.v0_1_0.TendermintMessage)
    },
)
_sym_db.RegisterMessage(TendermintMessage)
_sym_db.RegisterMessage(TendermintMessage.ErrorCode)
_sym_db.RegisterMessage(TendermintMessage.Get_Genesis_Info_Performative)
_sym_db.RegisterMessage(TendermintMessage.Get_Recovery_Params_Performative)
_sym_db.RegisterMessage(TendermintMessage.Genesis_Info_Performative)
_sym_db.RegisterMessage(TendermintMessage.Recovery_Params_Performative)
_sym_db.RegisterMessage(TendermintMessage.Error_Performative)
_sym_db.RegisterMessage(TendermintMessage.Error_Performative.ErrorDataEntry)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE_ERRORDATAENTRY._options = None
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE_ERRORDATAENTRY._serialized_options = b"8\001"
    _TENDERMINTMESSAGE._serialized_start = 51
    _TENDERMINTMESSAGE._serialized_end = 1248
    _TENDERMINTMESSAGE_ERRORCODE._serialized_start = 582
    _TENDERMINTMESSAGE_ERRORCODE._serialized_end = 724
    _TENDERMINTMESSAGE_ERRORCODE_ERRORCODEENUM._serialized_start = 688
    _TENDERMINTMESSAGE_ERRORCODE_ERRORCODEENUM._serialized_end = 724
    _TENDERMINTMESSAGE_GET_GENESIS_INFO_PERFORMATIVE._serialized_start = 726
    _TENDERMINTMESSAGE_GET_GENESIS_INFO_PERFORMATIVE._serialized_end = 794
    _TENDERMINTMESSAGE_GET_RECOVERY_PARAMS_PERFORMATIVE._serialized_start = 796
    _TENDERMINTMESSAGE_GET_RECOVERY_PARAMS_PERFORMATIVE._serialized_end = 867
    _TENDERMINTMESSAGE_GENESIS_INFO_PERFORMATIVE._serialized_start = 869
    _TENDERMINTMESSAGE_GENESIS_INFO_PERFORMATIVE._serialized_end = 910
    _TENDERMINTMESSAGE_RECOVERY_PARAMS_PERFORMATIVE._serialized_start = 912
    _TENDERMINTMESSAGE_RECOVERY_PARAMS_PERFORMATIVE._serialized_end = 958
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE._serialized_start = 961
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE._serialized_end = 1232
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE_ERRORDATAENTRY._serialized_start = 1184
    _TENDERMINTMESSAGE_ERROR_PERFORMATIVE_ERRORDATAENTRY._serialized_end = 1232
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Constants"""
import os

from autonomy.__version__ import __version__ as DEFAULT_AUTONOMY_VERSION


DEFAULT_BUILD_FOLDER = "abci_build"
DEFAULT_KEYS_FILE = "keys.json"
DEFAULT_IMAGE_VERSION = "latest"
SERVICE_REGISTRY_CONTRACT_CONTAINER_NAME = "autonolas-registries"
DOCKER_COMPOSE_YAML = "docker-compose.yaml"
VALORY = "valory"

AUTONOMY_IMAGE_VERSION = os.environ.get(
    "AUTONOMY_IMAGE_VERSION", DEFAULT_AUTONOMY_VERSION
)
TENDERMINT_IMAGE_VERSION = os.environ.get(
    "TENDERMINT_IMAGE_VERSION", DEFAULT_AUTONOMY_VERSION
)
HARDHAT_IMAGE_VERSION = os.environ.get("HARDHAT_IMAGE_VERSION", DEFAULT_IMAGE_VERSION)
SERVICE_REGISTRY_IMAGE_VERSION = os.environ.get(
    "SERVICE_REGISTRY_IMAGE_VERSION", DEFAULT_IMAGE_VERSION
)
ACN_IMAGE_VERSION = os.environ.get("ACN_IMAGE_VERSION", DEFAULT_IMAGE_VERSION)

AUTONOMY_IMAGE_NAME = os.environ.get("AUTONOMY_IMAGE_NAME", "valory/open-autonomy")
TENDERMINT_IMAGE_NAME = os.environ.get(
    "TENDERMINT_IMAGE_NAME", "valory/open-autonomy-tendermint"
)
HARDHAT_IMAGE_NAME = os.environ.get(
    "HARDHAT_IMAGE_NAME", "valory/open-autonomy-hardhat"
)
SERVICE_REGISTRY_IMAGE_NAME = os.environ.get(
    "SERVICE_REGISTRY_IMAGE_NAME", f"valory/{SERVICE_REGISTRY_CONTRACT_CONTAINER_NAME}"
)

DEFAULT_SERVICE_REGISTRY_CONTRACTS_IMAGE = (
    f"{SERVICE_REGISTRY_IMAGE_NAME}:{SERVICE_REGISTRY_IMAGE_VERSION}"
)
ACN_IMAGE_NAME = os.environ.get("ACN_IMAGE_NAME", "valory/open-acn-node")
DEFAULT_DOCKER_IMAGE_AUTHOR = "valory"
OAR_IMAGE = "{image_author}/oar-{agent}:{version}"
ABSTRACT_ROUND_ABCI_SKILL_WITH_HASH = "valory/abstract_round_abci:0.1.0:bafybeibdljcw2wchyderkonobpdpyb36yioagqcaoqldydw5bi6vvempma"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Implementation of the 'autonomy publish' subcommand."""

from pathlib import Path

import click
from aea.cli.publish import publish_agent_package
from aea.cli.utils.click_utils import registry_flag, reraise_as_click_exception
from aea.configurations.constants import (
    DEFAULT_AEA_CONFIG_FILE,
    DEFAULT_SERVICE_CONFIG_FILE,
)

from autonomy.cli.helpers.registry import publish_service_package


@click.command(name="publish")
@registry_flag()
@click.option(
    "--push-missing", is_flag=True, help="Push missing components on the registry."
)
@click.pass_context
def publish(
    click_context: click.Context, registry: str, push_missing: bool
) -> None:  # pylint: disable=unused-argument
    """Publish the agent or service on the registry."""

    with reraise_as_click_exception(Exception):
        if Path(click_context.obj.cwd, DEFAULT_SERVICE_CONFIG_FILE).exists():
            # TODO: support push_missing
            publish_service_package(click_context, registry)
        elif Path(
            click_context.obj.cwd, DEFAULT_AEA_CONFIG_FILE
        ).exists():  # pragma: nocover
            publish_agent_package(click_context, registry, push_missing)
        else:
            raise FileNotFoundError(
                "No package config found in this directory."
            )  # pragma: no cover



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test `valory/open-autonomy` image."""

from pathlib import Path
from typing import cast

import docker
from docker.models.containers import Container

from autonomy.constants import AUTONOMY_IMAGE_NAME
from autonomy.deploy.constants import DOCKERFILES

from tests.conftest import ROOT_DIR, skip_docker_tests
from tests.test_autonomy.test_images.base import BaseImageBuildTest


@skip_docker_tests
class TestOpenAutonomyBaseImage(BaseImageBuildTest):
    """Test image build and run."""

    client: docker.DockerClient
    path: Path = ROOT_DIR / "deployments" / DOCKERFILES / "autonomy"
    tag: str = f"{AUTONOMY_IMAGE_NAME}:latest"

    def test_image(self) -> None:
        """Test image build."""

        success, output = self.build_image(
            path=self.path,
            tag=self.tag,
        )

        assert success, output
        assert "Successfully built" in output

        container = cast(
            Container,
            self.client.containers.run(
                image=self.tag,
                command=["-c", "ls"],
                detach=True,
            ),
        )

        container.wait()

        assert b"requirements.txt" in container.logs()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""
Implement a scaffold sub-command to scaffold ABCI skills.

This module patches the 'aea scaffold' command so to add a new subcommand for scaffolding a skill
 starting from FSM specification.
"""


from pathlib import Path

import click
from aea.cli.add import add_item
from aea.cli.scaffold import scaffold, scaffold_item
from aea.cli.utils.click_utils import registry_flag
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import pass_ctx
from aea.configurations.constants import SKILL

# the decoration does side-effect on the 'aea scaffold' command
from aea.configurations.data_types import PackageId, PackageType, PublicId
from aea.package_manager.v1 import PackageManagerV1

from autonomy.constants import ABSTRACT_ROUND_ABCI_SKILL_WITH_HASH
from autonomy.fsm.scaffold.scaffold_skill import (
    ScaffoldABCISkill,
    SkillConfigUpdater,
    TO_LOCAL_REGISTRY_FLAG,
)


def _add_abstract_round_abci_if_not_present(ctx: Context) -> None:
    """Add 'abstract_round_abci' skill if not present."""
    MSG = (
        "Skill valory/abstract_round_abci not found in agent dependencies, adding it..."
    )

    if ctx.config.get(TO_LOCAL_REGISTRY_FLAG):
        skill_id = SkillConfigUpdater.get_actual_abstract_round_abci_package_public_id(
            ctx
        )

        if not skill_id:
            # add package to local registry
            click.echo(MSG)
            package_manager = PackageManagerV1.from_dir(Path(ctx.registry_path))
            package_id = PackageId(
                package_type=PackageType.SKILL,
                public_id=PublicId.from_str(ABSTRACT_ROUND_ABCI_SKILL_WITH_HASH),
            )
            package_manager.add_package(package_id, with_dependencies=True)
            package_manager.dump()
    else:
        # add package to agent
        abstract_round_abci_public_id = PublicId.from_str(
            ABSTRACT_ROUND_ABCI_SKILL_WITH_HASH
        )
        if abstract_round_abci_public_id.to_latest() not in {
            public_id.to_latest() for public_id in ctx.agent_config.skills
        }:
            click.echo(MSG)

            add_item(ctx, SKILL, abstract_round_abci_public_id)


@scaffold.command()  # noqa
@registry_flag()
@click.argument("skill_name", type=str, required=True)
@click.option("--spec", type=click.Path(exists=True, dir_okay=False), required=True)
@pass_ctx
def fsm(ctx: Context, registry: str, skill_name: str, spec: str) -> None:
    """Add an ABCI skill scaffolding from an FSM specification."""
    ctx.registry_type = registry

    # check abstract_round_abci is in dependencies; if not, add it
    _add_abstract_round_abci_if_not_present(ctx)

    # scaffold AEA skill - as usual
    # TODO: open-aea/aea/cli/scaffold.py:279 bad trick with update of ctx.cwd
    preserve_cwd = ctx.cwd
    scaffold_item(ctx, SKILL, skill_name)
    ctx.cwd = preserve_cwd

    scaffold_fsm = ScaffoldABCISkill(ctx, skill_name, Path(spec))
    scaffold_fsm.do_scaffolding()

    if ctx.config[TO_LOCAL_REGISTRY_FLAG]:
        scaffold_fsm.add_skill_to_packages()
